-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
--Mandatory controller parameters
M.type = "main"
M.relevantDevice = "mainEngine"
M.engineInfo = {}
M.fireEngineTemperature = 0
M.throttle = 0
M.brake = 0
M.clutchRatio = 0
-----

local settings = require("simplesettings")

local min = math.min
local max = math.max
local abs = math.abs

local constants = {rpmToAV = 0.104719755, avToRPM = 9.549296596425384}

local expectedDeviceNames = {engine = "mainEngine", gearbox = "gearbox", torqueConverter = "torqueConverter"}

local gearboxHandling = {
  behaviors = {"arcade", "realistic"},
  behaviorLookup = nil,
  behavior = nil,
  previousBehavior = nil,
  logic = nil,
  autoClutch = true,
  autoThrottle = true,
  autoShiftPrevention = true,
  availableLogic = {}
}

local automaticHandling = {
  availableModes = {"P","R", "N", "D", "S", "1", "2", "M"},
  hShifterModeLookup = {[-1] = "R", [0] = "N", "P", "D", "S", "2", "1", "M1"},
  cvtGearIndexLookup = {P = -2, R = -1, N = 0, D = 1, S = 2, ["2"] = 3, ["1"]= 4, M1 = 5},
  availableModeLookup = {},
  existingModeLookup = {},
  modeIndexLookup = {},
  modes = {},
  mode = nil,
  modeIndex = 0,
  maxAllowedGearIndex = 0,
  minAllowedGearIndex = 0,
  arcadeAutoBrakeAmount = 0.2,
  arcadeAutoBrakeAVThreshold = 5,
}

local clutchHandling = {
  clutchInRate = 5,
  clutchOutRate = 5,
  clutchLaunchTargetAV = 0,
  clutchLaunchStartAV = 0,
  clutchLaunchIFactor = 0,
  preShiftClutchRatio = 0,
  shiftState = "clutchIn",
  revMatchThrottle = 0.5,
  didRevMatch = false,
}

local dct = {
  access1 = {clutchRatioName= "clutchRatio1", gearIndexName = "gearIndex1", setGearIndexName = "setGearIndex1"},
  access2 = {clutchRatioName= "clutchRatio2", gearIndexName = "gearIndex2", setGearIndexName = "setGearIndex2"},
  primaryAccess = nil,
  secondaryAccess = nil,
  clutchTime = 0,
}

local cvtHandling = {
  aggression = 0.5,
  highAV = 0,
  lowAV = 0,
}

local energyStorageData = {
  ratio = 0,
  volume = 0,
  capacity = 0,
  invEnergyStorageCount = 0,
}

local shiftPreventionData = {
  wheelSlipUpThreshold = 0,
  wheelSlipDownThreshold = 0,
  wheelSlipShiftDown = false,
  wheelSlipShiftUp = false
}

local shiftBehavior = {
  shiftUpAV = 0,
  shiftDownAV = 0,
}

local torqueConverterHandling = {
  lockupAV = 0,
  lockupRange = 0,
  lockupMinGear = 0,
  hasLockup = false,
}

local smoother = {
  throttleSmoother = nil,
  avgAVsmoother = nil,
  --gearRatio smoother reduces oscillations during sharp changes in throttle
  cvtGearRatioSmoother = nil,
  --target smoother represents "driver intent"
  cvtTargetAVSmoother = nil,
  wheelSlipShiftUpSmoother = nil,
}

local engine = nil
local gearbox = nil
local gearboxType = nil

local timerConstants = {gearChangeDelay = 0, shiftDelay = 0, neutralSelectionDelay = 0}
local timer = {gearChangeDelayTimer = 0, shiftDelayTimer = 0, neutralSelectionDelayTimer = 0, stalledEngineMessageTimer = 0}

local shiftPoints = {}
local currentGearIndex = 0
local newDesiredGearIndex = 0
local previousGearIndex = 0
local updateGearbox = nil
local isStartingEngine = false
local engineWasStalled = false
local engineManuallyTurnedOff = false
--local playWheelShiftLEDs = false

local inputValues = {throttle = 0, clutch = 0}

local isFrozen = false

--Used for decision logic, smoothed throttle is not actually used as the throttle value
local smoothedThrottle = 0
local smoothThrottleUpShiftThreshold = 0.05
local smoothedAvgAV = 0

--------------------------------------
--------------- Shared ---------------
--------------------------------------

local sharedFunctions = {}

sharedFunctions.selectShiftPoints = function(gearIndex, sportThrottleOverride)
  --interpolate based on throttle between high/low ranges
  local throttleCubed = sportThrottleOverride and 1 or (M.throttle * M.throttle * M.throttle)
  local shiftPoint = shiftPoints[gearIndex]
  shiftBehavior.shiftDownAV = shiftPoint.lowShiftDownAV + (shiftPoint.highShiftDownAV - shiftPoint.lowShiftDownAV) * throttleCubed
  shiftBehavior.shiftUpAV = shiftPoint.lowShiftUpAV + (shiftPoint.highShiftUpAV - shiftPoint.lowShiftUpAV) * throttleCubed
end

sharedFunctions.applyGearboxModeRestrictions = function()
  local manualModeIndex
  if string.sub(automaticHandling.mode, 1,1) == "M" then
    manualModeIndex = string.sub(automaticHandling.mode, 2)
  end
  local maxGearIndex = gearbox.maxGearIndex
  local minGearIndex = gearbox.minGearIndex
  if automaticHandling.mode == "1" then
    maxGearIndex = 1
    minGearIndex = 1
  elseif automaticHandling.mode == "2" then
    maxGearIndex = 2
    minGearIndex = 1
  elseif manualModeIndex then
    maxGearIndex = manualModeIndex
    minGearIndex = manualModeIndex
  end

  automaticHandling.maxGearIndex = maxGearIndex
  automaticHandling.minGearIndex = minGearIndex
end

local test = obj:queueGameEngineLua("be:getPlayerVehicle(0):queueLuaCommand('sounds.cameraMode = (\"'..camera.getActiveMode(0)..'\")')") --//////////////////
local isOn = 0																															 --//////////////////
--------------------------------------
--------------- Manual ---------------
--------------------------------------

local manualFunctions = {}

manualFunctions.manualGearboxBehaviorChanged = function()
  if gearboxHandling.behavior == "realistic" and not gearboxHandling.autoClutch and abs(gearbox.gearIndex) == 1 then
    gearbox:setGearIndex(0)
  end
end

manualFunctions.shiftUpManual = function()
  local gearIndex = newDesiredGearIndex ~= 0 and newDesiredGearIndex or gearbox.gearIndex
  gearIndex = min(max(gearIndex + 1, gearbox.minGearIndex), gearbox.maxGearIndex)
  if gearbox.gearIndex ~= gearIndex then
	test = obj:queueGameEngineLua("be:getPlayerVehicle(0):queueLuaCommand('sounds.cameraMode = (\"'..camera.getActiveMode(0)..'\")')") --//////////////////
	if sounds.cameraMode == "dash" then																							   	   --//////////////////
		alea = math.random (3)  																									   --//////////////////
		if alea == 1 then 																											   --//////////////////
		sounds.playSoundOnceAtNode("shift1", v.data.refNodes[0].ref, 1)															       --//////////////////
		elseif alea == 2 then																										   --//////////////////
		sounds.playSoundOnceAtNode("shift2", v.data.refNodes[0].ref, 1)															       --//////////////////
		else																														   --//////////////////
		sounds.playSoundOnceAtNode("shift3", v.data.refNodes[0].ref, 1)															       --//////////////////
		end																															   --//////////////////
	end																																   --//////////////////
	
	
	
    newDesiredGearIndex = gearIndex
    previousGearIndex = gearbox.gearIndex
    clutchHandling.shiftState = "clutchIn"
    clutchHandling.preShiftClutchRatio = M.clutchRatio
    updateGearbox = gearboxHandling.logic.whileShifting
  end
end

manualFunctions.shiftDownManual = function()
  local gearIndex = newDesiredGearIndex ~= 0 and newDesiredGearIndex or gearbox.gearIndex
  gearIndex = min(max(gearIndex - 1, gearbox.minGearIndex), gearbox.maxGearIndex)
  if gearbox.gearIndex ~= gearIndex then
	test = obj:queueGameEngineLua("be:getPlayerVehicle(0):queueLuaCommand('sounds.cameraMode = (\"'..camera.getActiveMode(0)..'\")')") --//////////////////
	if sounds.cameraMode == "dash" then																							   	   --//////////////////
		alea1 = math.random (3)  																									   --//////////////////
		if alea1 == 1 then 																											   --//////////////////
		sounds.playSoundOnceAtNode("shift1", v.data.refNodes[0].ref, 1)															       --//////////////////
		elseif alea1 == 2 then																										   --//////////////////
		sounds.playSoundOnceAtNode("shift2", v.data.refNodes[0].ref, 1)															       --//////////////////
		else																														   --//////////////////
		sounds.playSoundOnceAtNode("shift3", v.data.refNodes[0].ref, 1)															       --//////////////////
		end																															   --//////////////////
	end																																   --//////////////////
    newDesiredGearIndex = gearIndex
    previousGearIndex = gearbox.gearIndex
    clutchHandling.shiftState = "clutchIn"
    clutchHandling.preShiftClutchRatio = M.clutchRatio
    updateGearbox = gearboxHandling.logic.whileShifting
  end
end

manualFunctions.shiftToGearIndexManual = function(index)
  local gearIndex = min(max(index, gearbox.minGearIndex), gearbox.maxGearIndex)
  if gearbox.gearIndex ~= gearIndex then
    print("???")
    newDesiredGearIndex = gearIndex
    previousGearIndex = gearbox.gearIndex
    timer.shiftDelayTimer = 0
    clutchHandling.shiftState = "clutchIn"
    clutchHandling.preShiftClutchRatio = M.clutchRatio
    updateGearbox = gearboxHandling.logic.whileShifting
  end
end

manualFunctions.updateManualInGearArcade = function(dt)
  M.throttle = inputValues.throttle
  local gearIndex = gearbox.gearIndex
  -- driving backwards? - only with automatic shift - for obvious reasons ;)
  if (gearIndex < 0 and smoothedAvgAV <= 0.15) or (gearIndex <= 0 and smoothedAvgAV < -1) then
    M.throttle, M.brake = M.brake, M.throttle
  end

  local engineAV = engine.outputAV1
  if timer.gearChangeDelayTimer <= 0 and gearIndex ~= 0 then
    local tmpEngineAV = engineAV
    local relEngineAV = engineAV / gearbox.gearRatio

    sharedFunctions.selectShiftPoints(gearIndex)

    --shift down?
    while tmpEngineAV < shiftBehavior.shiftDownAV and abs(gearIndex) > 1 and shiftPreventionData.wheelSlipShiftDown and abs(M.throttle - smoothedThrottle) < smoothThrottleUpShiftThreshold do
      gearIndex = gearIndex - fsign(gearIndex)
      tmpEngineAV = relEngineAV * (gearbox.gearRatios[gearIndex] or 0)
      if tmpEngineAV > engine.maxAV then
        gearIndex = gearIndex + fsign(gearIndex)
        break
      end
      sharedFunctions.selectShiftPoints(gearIndex)
    end

    --shift up?
    if M.clutchRatio >= 1 and tmpEngineAV >= shiftBehavior.shiftUpAV and shiftPreventionData.wheelSlipShiftUp and M.brake <= 0 and abs(M.throttle - smoothedThrottle) < smoothThrottleUpShiftThreshold and gearIndex < gearbox.maxGearIndex and gearIndex > gearbox.minGearIndex then
      gearIndex = gearIndex + fsign(gearIndex)
      tmpEngineAV = relEngineAV * (gearbox.gearRatios[gearIndex] or 0)
      if tmpEngineAV < engine.idleAV then
        gearIndex = gearIndex - fsign(gearIndex)
      end
      sharedFunctions.selectShiftPoints(gearIndex)
    end
  end

  -- neutral gear handling
  if abs(gearIndex) <= 1 and timer.neutralSelectionDelayTimer <= 0 then
    if gearIndex ~= 0 and abs(smoothedAvgAV) < automaticHandling.arcadeAutoBrakeAVThreshold and M.throttle <= 0 then
      M.brake = max(M.brake, automaticHandling.arcadeAutoBrakeAmount)
    end

    if inputValues.throttle > 0 and smoothedAvgAV > -1 and gearIndex < 1 then
      gearIndex = 1
      timer.neutralSelectionDelayTimer = timerConstants.neutralSelectionDelay
    end

    if input.brake > 0 and inputValues.throttle <= 0 and smoothedAvgAV <= 0.15 and gearIndex > -1 then
      gearIndex = -1
      timer.neutralSelectionDelayTimer = timerConstants.neutralSelectionDelay
    end
  end

  if gearbox.gearIndex ~= gearIndex then
	test = obj:queueGameEngineLua("be:getPlayerVehicle(0):queueLuaCommand('sounds.cameraMode = (\"'..camera.getActiveMode(0)..'\")')") --//////////////////
	if sounds.cameraMode == "dash" then																							   	   --//////////////////
		alea2 = math.random (3)  																									   --//////////////////
		if alea2 == 1 then 																											   --//////////////////
		sounds.playSoundOnceAtNode("shift1", v.data.refNodes[0].ref, 1)															       --//////////////////
		elseif alea2 == 2 then																										   --//////////////////
		sounds.playSoundOnceAtNode("shift2", v.data.refNodes[0].ref, 1)															       --//////////////////
		else																														   --//////////////////
		sounds.playSoundOnceAtNode("shift3", v.data.refNodes[0].ref, 1)															       --//////////////////
		end																															   --//////////////////
	end																																   --//////////////////
    newDesiredGearIndex = gearIndex
    previousGearIndex = gearbox.gearIndex
    clutchHandling.shiftState = "clutchIn"
    updateGearbox = gearboxHandling.logic.whileShifting
  end

  -- Control clutch to buildup engine RPM
  if abs(gearIndex) == 1 and M.throttle > 0 then
    local ratio = max((engine.outputAV1 - clutchHandling.clutchLaunchStartAV * (1 + M.throttle)) / (clutchHandling.clutchLaunchTargetAV * (1 + clutchHandling.clutchLaunchIFactor)), 0)
    clutchHandling.clutchLaunchIFactor = min(clutchHandling.clutchLaunchIFactor + dt * 0.5, 1)
    M.clutchRatio = math.min(math.max(ratio * ratio, 0), 1)
  else
    clutchHandling.clutchLaunchIFactor = 0
    M.clutchRatio = 1
  end

  if inputValues.clutch > 0 then
    M.clutchRatio = min(1 - inputValues.clutch, M.clutchRatio)
    timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
  end

  --always prevent stalling
  local stallPrevent = min(max((engine.outputAV1 * 0.95 - engine.idleAV ) / (engine.idleAV * 0.1), 0), 1)
  M.clutchRatio = min(M.clutchRatio, stallPrevent * stallPrevent)

  clutchHandling.preShiftClutchRatio = M.clutchRatio
  currentGearIndex = gearIndex
end

manualFunctions.updateManualWhileShiftingArcade = function(dt)
  -- old -> N -> wait -> new -> in gear update
  local gearIndex = gearbox.gearIndex
  if (gearIndex < 0 and smoothedAvgAV <= 0.15) or (gearIndex <= 0 and smoothedAvgAV < -1) then
    M.throttle, M.brake = M.brake, M.throttle
  end

  M.throttle = 0
  if clutchHandling.shiftState == "clutchIn" then
    M.clutchRatio = max(M.clutchRatio - dt * clutchHandling.clutchInRate, 0)
    if M.clutchRatio <= 0 then
      if previousGearIndex ~= 0 then
        clutchHandling.shiftState = "neutral"
      else
        clutchHandling.shiftState = "shift"
      end
    end

  elseif clutchHandling.shiftState == "neutral" then
    gearbox:setGearIndex(0)
    timer.shiftDelayTimer = timerConstants.shiftDelay
    clutchHandling.shiftState = "shift"

  elseif clutchHandling.shiftState == "shift" then
    local canShift = true
    local targetAV = (gearbox.gearRatios[newDesiredGearIndex] / gearbox.gearRatios[previousGearIndex]) * (gearbox.outputAV1 * gearbox.gearRatios[previousGearIndex])
    if targetAV > engine.outputAV1 and previousGearIndex ~= 0 and not clutchHandling.didRevMatch and clutchHandling.preShiftClutchRatio >= 1 then
      M.throttle = clutchHandling.revMatchThrottle
      canShift = engine.outputAV1 >= targetAV or targetAV > engine.maxAV
      clutchHandling.didRevMatch = canShift
    end
    if timer.shiftDelayTimer <= 0 and canShift then
      gearbox:setGearIndex(newDesiredGearIndex)
      newDesiredGearIndex = 0
      previousGearIndex = 0
      timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
      clutchHandling.didRevMatch = false
      clutchHandling.shiftState = "clutchOut"
    end

  elseif clutchHandling.shiftState == "clutchOut" then
    if clutchHandling.preShiftClutchRatio > 0 then
      local stallPrevent = min(max((engine.outputAV1 * 0.9 - engine.idleAV) / (engine.idleAV * 0.1), 0), 1)
      M.clutchRatio = min(M.clutchRatio + dt * clutchHandling.clutchOutRate, stallPrevent * stallPrevent)
      if M.clutchRatio >= 1 or stallPrevent < 1 then
        updateGearbox = gearboxHandling.logic.inGear
      end
    else
      updateGearbox = gearboxHandling.logic.inGear
    end
  end
end

manualFunctions.updateManualInGear = function(dt)
  M.throttle = inputValues.throttle
  -- Control clutch to buildup engine RPM
  if gearboxHandling.autoClutch then
    if abs(gearbox.gearIndex) == 1 and M.throttle > 0 then
      local ratio = max((engine.outputAV1 - clutchHandling.clutchLaunchStartAV * (1 + M.throttle)) / (clutchHandling.clutchLaunchTargetAV * (1 + clutchHandling.clutchLaunchIFactor)), 0)
      clutchHandling.clutchLaunchIFactor = min(clutchHandling.clutchLaunchIFactor + dt * 0.5, 1)
      M.clutchRatio = math.min(math.max(ratio * ratio, 0), 1)
    else
      clutchHandling.clutchLaunchIFactor = 0
      M.clutchRatio = 1
    end

    if inputValues.clutch > 0 then
      M.clutchRatio = min(1 - inputValues.clutch, M.clutchRatio)
    end

    --always prevent stalling
    local stallPrevent = min(max((engine.outputAV1 * 0.95 - engine.idleAV) / (engine.idleAV * 0.1), 0), 1)
    M.clutchRatio = min(M.clutchRatio, stallPrevent * stallPrevent)
  else
    M.clutchRatio = 1 - inputValues.clutch
  end
  currentGearIndex = gearbox.gearIndex
end

manualFunctions.updateManualWhileShifting = function(dt)
  -- old -> N -> wait -> new -> in gear update
  if gearboxHandling.autoThrottle then
    M.throttle = 0
  else
    M.throttle = inputValues.throttle
  end

  if gearboxHandling.autoClutch then
    if clutchHandling.shiftState == "clutchIn" then
      M.clutchRatio = max(M.clutchRatio - dt * clutchHandling.clutchInRate, 0)
      if M.clutchRatio <= 0 then
        if previousGearIndex ~= 0 then
          clutchHandling.shiftState = "neutral"
        else
          clutchHandling.shiftState = "shift"
        end
      end

    elseif clutchHandling.shiftState == "neutral" then
      gearbox:setGearIndex(0)
      timer.shiftDelayTimer = timerConstants.shiftDelay
      clutchHandling.shiftState = "shift"

    elseif clutchHandling.shiftState == "shift" then
      local canShift = true
      local targetAV = gearbox.gearRatios[newDesiredGearIndex] * gearbox.outputAV1
      if gearboxHandling.autoThrottle and targetAV > engine.outputAV1 and previousGearIndex ~= 0 and not clutchHandling.didRevMatch and clutchHandling.preShiftClutchRatio >= 1 then
        M.throttle = clutchHandling.revMatchThrottle
        canShift = engine.outputAV1 >= targetAV or targetAV > engine.maxAV
        clutchHandling.didRevMatch = canShift
      end
      if timer.shiftDelayTimer <= 0 and canShift then
        gearbox:setGearIndex(newDesiredGearIndex)
        newDesiredGearIndex = 0
        previousGearIndex = 0
        timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
        clutchHandling.didRevMatch = false
        clutchHandling.shiftState = "clutchOut"
      end
    elseif clutchHandling.shiftState == "clutchOut" then
      if clutchHandling.preShiftClutchRatio > 0 then
        local stallPrevent = min(max((engine.outputAV1 * 0.9 - engine.idleAV) / (engine.idleAV * 0.1), 0), 1)
        M.clutchRatio = min(M.clutchRatio + dt * clutchHandling.clutchOutRate, stallPrevent)
        if M.clutchRatio >= 1 or stallPrevent < 1 then
          updateGearbox = gearboxHandling.logic.inGear
        end
      else
        updateGearbox = gearboxHandling.logic.inGear
      end
    end

  else
    M.clutchRatio = 1 - inputValues.clutch
    if timer.shiftDelayTimer <= 0 then
      gearbox:setGearIndex(newDesiredGearIndex)
      newDesiredGearIndex = 0
      previousGearIndex = 0
      timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
      updateGearbox = gearboxHandling.logic.inGear
    end
  end
end


--------------------------------------
------------- Automatic --------------
--------------------------------------

local automaticFunctions = {}

automaticFunctions.applyAutoGearboxMode = function()
  local autoIndex = automaticHandling.modeIndexLookup[automaticHandling.mode]
  if autoIndex then
    automaticHandling.modeIndex = min(max(autoIndex, 1), #automaticHandling.modes)
    automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]
  end

  if automaticHandling.mode == "P" then
    gearbox:setGearIndex(0)
    gearbox:setMode("park")
  elseif automaticHandling.mode == "N" then
    gearbox:setGearIndex(0)
    gearbox:setMode("neutral")
  else
    gearbox:setMode("drive")
    if automaticHandling.mode == "R" and gearbox.gearIndex > -1 then
      gearbox:setGearIndex(-1)
    elseif automaticHandling.mode ~= "R" and gearbox.gearIndex < 1 then
      gearbox:setGearIndex(1)
    end
  end
end

automaticFunctions.shiftUpAutomatic = function()
  if automaticHandling.mode == "N" then
    timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
  end

  automaticHandling.modeIndex = min(automaticHandling.modeIndex + 1, #automaticHandling.modes)
  automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]

  if automaticHandling.mode == "M1" then --we just shifted into M1
    automaticHandling.mode = "M"..tostring(max(gearbox.gearIndex, 1))
  end

  automaticFunctions.applyAutoGearboxMode()
  sharedFunctions.applyGearboxModeRestrictions()
end

automaticFunctions.shiftDownAutomatic = function()
  if automaticHandling.mode == "N" then
    timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
  end

  local previousMode = automaticHandling.mode
  automaticHandling.modeIndex = max(automaticHandling.modeIndex - 1, 1)
  automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]

  if previousMode == "M1" and electrics.values.wheelspeed > 2 and gearboxHandling.autoShiftPrevention then
    --we just tried to downshift past M1, something that is irritating while racing, so we disallow this shift unless we are really slow
    automaticHandling.mode = previousMode
  end

  --  if string.find(previousMode, "M") and gearboxHandling.autoShiftPrevention then
  --    local gearIndex = string.sub(automaticHandling.mode, 2)
  --    local gearRatio = gearbox.gearRatios[gearIndex]

  --  end

  automaticFunctions.applyAutoGearboxMode()
  sharedFunctions.applyGearboxModeRestrictions()
end

automaticFunctions.shiftToGearIndexAutomatic = function(index)
  local desiredMode = automaticHandling.hShifterModeLookup[index]
  if not desiredMode or not automaticHandling.existingModeLookup[desiredMode] then
    if desiredMode and not automaticHandling.existingModeLookup[desiredMode] then
      gui.message({txt = "vehicle.drivetrain.cannotShiftAuto", context = {mode = desiredMode}})
    end
    desiredMode = "N"
  end
  automaticHandling.mode = desiredMode

  automaticFunctions.applyAutoGearboxMode()
  sharedFunctions.applyGearboxModeRestrictions()
end

automaticFunctions.updateAutomaticInGearArcade = function(dt)
  M.throttle = inputValues.throttle
  local gearIndex = gearbox.gearIndex
  -- driving backwards? - only with automatic shift - for obvious reasons ;)
  if (gearIndex < 0 and smoothedAvgAV <= 0.15) or (gearIndex <= 0 and smoothedAvgAV < -1) then
    M.throttle, M.brake = M.brake, M.throttle
  end

  local engineAV = engine.outputAV1
  local gearboxInputAV = gearbox.inputAV

  if timer.gearChangeDelayTimer <= 0 and automaticHandling.mode ~= "N" then
    local tmpEngineAV = gearboxInputAV
    local relEngineAV = gearboxInputAV / gearbox.gearRatio

    sharedFunctions.selectShiftPoints(gearIndex)

    while tmpEngineAV < shiftBehavior.shiftDownAV and abs(gearIndex) > 1 and shiftPreventionData.wheelSlipShiftDown and abs(M.throttle - smoothedThrottle) < smoothThrottleUpShiftThreshold do
      gearIndex = gearIndex - fsign(gearIndex)
      tmpEngineAV = relEngineAV * (gearbox.gearRatios[gearIndex] or 0)
      if tmpEngineAV > engine.maxAV then
        gearIndex = gearIndex + fsign(gearIndex)
        break
      end
      sharedFunctions.selectShiftPoints(gearIndex)
    end

    --shift up?
    if tmpEngineAV >= shiftBehavior.shiftUpAV and M.brake <= 0 and shiftPreventionData.wheelSlipShiftUp and abs(M.throttle - smoothedThrottle) < smoothThrottleUpShiftThreshold and gearIndex < gearbox.maxGearIndex and gearIndex > gearbox.minGearIndex then
      gearIndex = gearIndex + fsign(gearIndex)
      tmpEngineAV = relEngineAV * (gearbox.gearRatios[gearIndex] or 0)
      if tmpEngineAV < engine.idleAV then
        gearIndex = gearIndex - fsign(gearIndex)
      end
      sharedFunctions.selectShiftPoints(gearIndex)
    end
  end

  if torqueConverterHandling.hasLockup and gearIndex >= torqueConverterHandling.lockupMinGear then
    electrics.values.lockupClutchRatio = min(max((gearboxInputAV - torqueConverterHandling.lockupAV) / torqueConverterHandling.lockupRange, 0), 1)
  else
    electrics.values.lockupClutchRatio = 0
  end

  -- neutral gear handling
  if abs(gearbox.gearIndex) <= 1 and timer.neutralSelectionDelayTimer <= 0 then
    if automaticHandling.mode ~= "P" and abs(smoothedAvgAV) < automaticHandling.arcadeAutoBrakeAVThreshold and M.throttle <= 0 then
      gearIndex = 0
      automaticHandling.mode = "N"
      automaticFunctions.applyAutoGearboxMode()
    end

    if inputValues.throttle > 0 and smoothedAvgAV > -1 and gearIndex < 1 then
      gearIndex = 1
      timer.neutralSelectionDelayTimer = timerConstants.neutralSelectionDelay
      automaticHandling.mode = "D"
      automaticFunctions.applyAutoGearboxMode()
    end

    if input.brake > 0 and inputValues.throttle <= 0 and smoothedAvgAV <= 0.15 and gearIndex > -1 then
      gearIndex = -1
      timer.neutralSelectionDelayTimer = timerConstants.neutralSelectionDelay
      automaticHandling.mode = "R"
      automaticFunctions.applyAutoGearboxMode()
    end
  end

  if gearbox.gearIndex ~= gearIndex then
    newDesiredGearIndex = gearIndex
    previousGearIndex = gearbox.gearIndex
    updateGearbox = gearboxHandling.logic.whileShifting
  end

  currentGearIndex = gearIndex
end

automaticFunctions.updateAutomaticWhileShiftingArcade = function()
  -- old -> wait -> new -> in gear update
  local gearIndex = gearbox.gearIndex
  if (gearIndex < 0 and smoothedAvgAV <= 0.15) or (gearIndex <= 0 and smoothedAvgAV < -1) then
    M.throttle, M.brake = M.brake, M.throttle
  end

  M.throttle = inputValues.throttle
  gearbox:setGearIndex(newDesiredGearIndex)
  newDesiredGearIndex = 0
  previousGearIndex = 0
  timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
  updateGearbox = gearboxHandling.logic.inGear
end

automaticFunctions.updateAutomaticInGear = function(dt)
  M.throttle = inputValues.throttle

  local gearIndex = gearbox.gearIndex
  local engineAV = engine.outputAV1
  local gearboxInputAV = gearbox.inputAV

  if timer.gearChangeDelayTimer <= 0 and automaticHandling.mode ~= "N" then
    local tmpEngineAV = gearboxInputAV
    local relEngineAV = gearboxInputAV / gearbox.gearRatio

    local isSportMode = automaticHandling.mode == "S"
    sharedFunctions.selectShiftPoints(gearIndex, isSportMode)

    --shift down?
    while tmpEngineAV < shiftBehavior.shiftDownAV and abs(gearIndex) > 1 and shiftPreventionData.wheelSlipShiftDown and abs(M.throttle - smoothedThrottle) < smoothThrottleUpShiftThreshold do
      gearIndex = gearIndex - fsign(gearIndex)
      tmpEngineAV = relEngineAV * (gearbox.gearRatios[gearIndex] or 0)
      if tmpEngineAV > engine.maxAV then
        gearIndex = gearIndex + fsign(gearIndex)
        break
      end
      sharedFunctions.selectShiftPoints(gearIndex, isSportMode)
    end

    --shift up?
    if tmpEngineAV >= shiftBehavior.shiftUpAV and M.brake <= 0 and shiftPreventionData.wheelSlipShiftUp and abs(M.throttle - smoothedThrottle) < smoothThrottleUpShiftThreshold and gearIndex < gearbox.maxGearIndex and gearIndex > gearbox.minGearIndex then
      gearIndex = gearIndex + fsign(gearIndex)
      tmpEngineAV = relEngineAV * (gearbox.gearRatios[gearIndex] or 0)
      if tmpEngineAV < engine.idleAV then
        gearIndex = gearIndex - fsign(gearIndex)
      end
      sharedFunctions.selectShiftPoints(gearIndex, isSportMode)
    end
  end

  --enforce things like L and M modes
  gearIndex = min(max(gearIndex, automaticHandling.minGearIndex), automaticHandling.maxGearIndex)

  if torqueConverterHandling.hasLockup and gearIndex >= torqueConverterHandling.lockupMinGear then
    electrics.values.lockupClutchRatio = min(max((gearboxInputAV - torqueConverterHandling.lockupAV) / torqueConverterHandling.lockupRange, 0), 1)
  else
    electrics.values.lockupClutchRatio = 0
  end

  if gearbox.gearIndex ~= gearIndex then
    newDesiredGearIndex = gearIndex
    previousGearIndex = gearbox.gearIndex
    updateGearbox = gearboxHandling.logic.whileShifting
  end

  currentGearIndex = gearIndex
end

automaticFunctions.updateAutomaticWhileShifting = function()
  -- old -> wait -> new -> in gear update
  M.throttle = inputValues.throttle
  gearbox:setGearIndex(newDesiredGearIndex)
  newDesiredGearIndex = 0
  previousGearIndex = 0
  timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
  updateGearbox = gearboxHandling.logic.inGear
end



--------------------------------------
----------------- DCT ----------------
--------------------------------------

local dctFunctions = {}

dctFunctions.applyDCTGearboxMode = function()
  local autoIndex = automaticHandling.modeIndexLookup[automaticHandling.mode]
  if autoIndex then
    automaticHandling.modeIndex = min(max(autoIndex, 1), #automaticHandling.modes)
    automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]
  end

  if automaticHandling.mode == "P" then
    gearbox:setMode("park")
  elseif automaticHandling.mode == "N" then
    gearbox:setMode("neutral")
  else
    gearbox:setMode("drive")
    local gearIndex = gearbox[dct.primaryAccess.gearIndexName]
    if automaticHandling.mode == "R" and gearbox[dct.access1.gearIndexName] > -1 then
      gearIndex = -1
      dct.primaryAccess = dct.access1
      dct.secondaryAccess = dct.access2
    elseif automaticHandling.mode ~= "R" and gearbox[dct.access1.gearIndexName] < 1 then
      gearIndex = 1
      dct.primaryAccess = dct.access1
      dct.secondaryAccess = dct.access2
    end

    if gearbox[dct.primaryAccess.gearIndexName] ~= gearIndex then
      newDesiredGearIndex = gearIndex
      previousGearIndex = gearbox[dct.primaryAccess.gearIndexName]
      timer.shiftDelayTimer = 0
      updateGearbox = gearboxHandling.logic.whileShifting
    end
  end
end

dctFunctions.shiftUpDCT = function()
  if automaticHandling.mode == "N" then
    timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
  end

  automaticHandling.modeIndex = min(automaticHandling.modeIndex + 1, #automaticHandling.modes)
  automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]

  if automaticHandling.mode == "M1" then --we just shifted into M1
    --instead of actually using M1, we want to KEEP the current gear, so M<current gear>
    automaticHandling.mode = "M"..tostring(max(gearbox.gearIndex, 1))
  end

  dctFunctions.applyDCTGearboxMode()
  sharedFunctions.applyGearboxModeRestrictions()
end

dctFunctions.shiftDownDCT = function()
  if automaticHandling.mode == "N" then
    timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
  end

  local previousMode = automaticHandling.mode
  automaticHandling.modeIndex = max(automaticHandling.modeIndex - 1, 1)
  automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]

  if previousMode == "M1" and electrics.values.wheelspeed > 2 and gearboxHandling.autoShiftPrevention then
    --we just tried to downshift past M1, something that is irritating while racing, so we disallow this shift unless we are really slow
    automaticHandling.mode = previousMode
  end

  dctFunctions.applyDCTGearboxMode()
  sharedFunctions.applyGearboxModeRestrictions()
end

dctFunctions.shiftToGearIndexDCT = function(index)
  local desiredMode = automaticHandling.hShifterModeLookup[index]
  if not desiredMode or not automaticHandling.existingModeLookup[desiredMode] then
    if desiredMode and not automaticHandling.existingModeLookup[desiredMode] then
      gui.message({txt = "vehicle.drivetrain.cannotShiftAuto", context = {mode = desiredMode}})
    end
    desiredMode = "N"
  end
  automaticHandling.mode = desiredMode

  dctFunctions.applyDCTGearboxMode()
  sharedFunctions.applyGearboxModeRestrictions()
end

dctFunctions.dctPredictNextGear = function()
  local nextGear = gearbox[dct.secondaryAccess.gearIndexName]
  if M.throttle > 0 and gearbox[dct.primaryAccess.gearIndexName] > 0 then
    nextGear = gearbox[dct.primaryAccess.gearIndexName] + fsign(gearbox[dct.primaryAccess.gearIndexName])
  elseif gearbox[dct.primaryAccess.gearIndexName] > 0 then
    nextGear = gearbox[dct.primaryAccess.gearIndexName] - fsign(gearbox[dct.primaryAccess.gearIndexName])
  end
  if gearbox[dct.secondaryAccess.gearIndexName] ~= nextGear and gearbox[dct.secondaryAccess.gearIndexName] < gearbox.maxGearIndex and timer.shiftDelayTimer <= 0 then
    gearbox[dct.secondaryAccess.setGearIndexName](gearbox, nextGear)
    timer.shiftDelayTimer = timerConstants.shiftDelay
  end
end

dctFunctions.updateDCTInGearArcade = function(dt)
  M.throttle = inputValues.throttle
  local gearIndex = gearbox[dct.primaryAccess.gearIndexName]
  -- driving backwards? - only with automatic shift - for obvious reasons ;)
  if (gearIndex < 0 and smoothedAvgAV <= 0.15) or (automaticHandling.mode == "N" and smoothedAvgAV < -1) then
    M.throttle, M.brake = M.brake, M.throttle
  end

  local engineAV = engine.outputAV1

  if timer.gearChangeDelayTimer <= 0 and automaticHandling.mode ~= "N" then

    local tmpEngineAV = engineAV
    local relEngineAV = engineAV / gearbox.gearRatios[gearIndex]

    sharedFunctions.selectShiftPoints(gearIndex)

    --shift down?
    while tmpEngineAV < shiftBehavior.shiftDownAV and abs(gearIndex) > 1 and shiftPreventionData.wheelSlipShiftDown and abs(M.throttle - smoothedThrottle) < smoothThrottleUpShiftThreshold do
      gearIndex = gearIndex - fsign(gearIndex)
      tmpEngineAV = relEngineAV * (gearbox.gearRatios[gearIndex] or 0)
      if tmpEngineAV > engine.maxAV then
        gearIndex = gearIndex + fsign(gearIndex)
        break
      end
      sharedFunctions.selectShiftPoints(gearIndex)
    end

    --shift up?
    if tmpEngineAV >= shiftBehavior.shiftUpAV and M.brake <= 0 and shiftPreventionData.wheelSlipShiftUp and abs(M.throttle - smoothedThrottle) < smoothThrottleUpShiftThreshold and gearIndex < gearbox.maxGearIndex and gearIndex > gearbox.minGearIndex then
      gearIndex = gearIndex + fsign(gearIndex)
      tmpEngineAV = relEngineAV * (gearbox.gearRatios[gearIndex] or 0)
      if tmpEngineAV < engine.idleAV then
        gearIndex = gearIndex - fsign(gearIndex)
      end
      sharedFunctions.selectShiftPoints(gearIndex)
    end
  end

  -- neutral gear handling
  if timer.neutralSelectionDelayTimer <= 0 then
    if automaticHandling.mode ~= "N" and automaticHandling.mode ~= "P" and abs(smoothedAvgAV) < automaticHandling.arcadeAutoBrakeAVThreshold and M.throttle <= 0 then
      M.brake = max(M.brake, automaticHandling.arcadeAutoBrakeAmount)
    end

    if inputValues.throttle > 0 and smoothedAvgAV > -1 and automaticHandling.mode ~= "D" then
      gearIndex = 1
      timer.neutralSelectionDelayTimer = timerConstants.neutralSelectionDelay
      automaticHandling.mode = "D"
      dctFunctions.applyDCTGearboxMode()
    end

    if input.brake > 0 and inputValues.throttle <= 0 and smoothedAvgAV <= 0.15 and automaticHandling.mode ~= "R" then
      gearIndex = -1
      timer.neutralSelectionDelayTimer = timerConstants.neutralSelectionDelay
      automaticHandling.mode = "R"
      dctFunctions.applyDCTGearboxMode()
    end
  end

  if gearbox[dct.primaryAccess.gearIndexName] ~= gearIndex then
    newDesiredGearIndex = gearIndex
    previousGearIndex = gearbox[dct.primaryAccess.gearIndexName]
    timer.shiftDelayTimer = 0
    updateGearbox = gearboxHandling.logic.whileShifting
  end

  -- Control clutch to buildup engine RPM
  local dctClutchRatio = 0
  if abs(gearIndex) == 1 and M.throttle > 0 then
    local ratio = max((engine.outputAV1 - clutchHandling.clutchLaunchStartAV * (1 + M.throttle)) / (clutchHandling.clutchLaunchTargetAV * (1 + clutchHandling.clutchLaunchIFactor)), 0)
    clutchHandling.clutchLaunchIFactor = min(clutchHandling.clutchLaunchIFactor + dt * 0.5, 1)
    dctClutchRatio = math.min(math.max(ratio * ratio, 0), 1)
  else
    clutchHandling.clutchLaunchIFactor = 0
    dctClutchRatio = 1
  end

  --always prevent stalling
  local stallPrevent = min(max((engine.outputAV1 * 0.95 - engine.idleAV) / (engine.idleAV * 0.1), 0), 1)
  dctClutchRatio = min(dctClutchRatio, stallPrevent * stallPrevent)

  electrics.values[dct.primaryAccess.clutchRatioName] = dctClutchRatio
  electrics.values[dct.secondaryAccess.clutchRatioName] = 0
  M.clutchRatio = 1
  currentGearIndex = (automaticHandling.mode == "N" or automaticHandling.mode == "P") and 0 or gearIndex
  gearbox.gearIndex = gearIndex --just so that the DCT can always present the "active" gear/ratio to the outside world
  gearbox.gearRatio = gearbox.gearRatios[gearIndex]

  dctFunctions.dctPredictNextGear()
end

dctFunctions.updateDCTWhileShiftingArcade = function(dt)
  local gearIndex = gearbox.gearIndex
  if (gearIndex < 0 and smoothedAvgAV <= 0.15) or (gearIndex <= 0 and smoothedAvgAV < -1) then
    M.throttle, M.brake = M.brake, M.throttle
  end

  M.throttle = inputValues.throttle
  -- secondary clutch closes while primary opens -> in gear update once fully closed
  local primaryGearIndex = gearbox[dct.primaryAccess.gearIndexName]
  local secondaryGearIndex = gearbox[dct.secondaryAccess.gearIndexName]
  if newDesiredGearIndex ~= secondaryGearIndex and timer.shiftDelayTimer <= 0 then
    --find out if our desired gear is actually on the secondary shaft
    local sameShaft = newDesiredGearIndex % 2 == secondaryGearIndex % 2
    --if so, we can directly shift to that desired gear on the secondary shaft
    --if not, we need to use a helper gear first which makes the actually desired gear part of the secondary shaft
    local newGearIndex = sameShaft and newDesiredGearIndex or primaryGearIndex + fsign(newDesiredGearIndex - primaryGearIndex)
    if secondaryGearIndex ~= newGearIndex then
      gearbox[dct.secondaryAccess.setGearIndexName](gearbox, newGearIndex)
      timer.shiftDelayTimer = timerConstants.shiftDelay
    end
  end

  local canShift = true
  local secondaryGearRatio = gearbox.gearRatios[secondaryGearIndex]
  local targetAV = secondaryGearRatio * gearbox.outputAV1
  if targetAV > engine.outputAV1 and not clutchHandling.didRevMatch then
    M.throttle = clutchHandling.revMatchThrottle
    electrics.values[dct.primaryAccess.clutchRatioName] = 0
    canShift = engine.outputAV1 >= targetAV or targetAV > engine.maxAV
    clutchHandling.didRevMatch = canShift
  end

  if timer.shiftDelayTimer <= 0 and canShift then
    local clutchRatio = min(electrics.values[dct.secondaryAccess.clutchRatioName] + (1 / dct.clutchTime) * dt, 1)
    local stallPrevent = min(max((engine.outputAV1 * 0.9 - engine.idleAV) / (engine.idleAV * 0.1), 0), 1)
    electrics.values[dct.primaryAccess.clutchRatioName] = min(1 - clutchRatio, stallPrevent * stallPrevent)
    electrics.values[dct.secondaryAccess.clutchRatioName] = min(clutchRatio, stallPrevent * stallPrevent)
    if clutchRatio == 1 or stallPrevent < 1 then
      dct.primaryAccess, dct.secondaryAccess = dct.secondaryAccess, dct.primaryAccess
      primaryGearIndex = gearbox[dct.primaryAccess.gearIndexName]

      if newDesiredGearIndex == primaryGearIndex then
        updateGearbox = gearboxHandling.logic.inGear
        timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
        clutchHandling.didRevMatch = false
      end
    end
  end

  gearbox.gearIndex = primaryGearIndex --just so that the DCT can always present the "active" gear to the outside world
end

dctFunctions.updateDCTInGear = function(dt)
  M.throttle = inputValues.throttle
  local gearIndex = gearbox[dct.primaryAccess.gearIndexName]

  local engineAV = engine.outputAV1

  if timer.gearChangeDelayTimer <= 0 and automaticHandling.mode ~= "N" then

    local tmpEngineAV = engineAV
    local relEngineAV = engineAV / gearbox.gearRatios[gearIndex]

    local isSportMode = automaticHandling.mode == "S"
    sharedFunctions.selectShiftPoints(gearIndex, isSportMode)

    while tmpEngineAV < shiftBehavior.shiftDownAV and abs(gearIndex) > 1 and shiftPreventionData.wheelSlipShiftDown and abs(M.throttle - smoothedThrottle) < smoothThrottleUpShiftThreshold do
      gearIndex = gearIndex - fsign(gearIndex)
      tmpEngineAV = relEngineAV * (gearbox.gearRatios[gearIndex] or 0)
      if tmpEngineAV > engine.maxAV then
        gearIndex = gearIndex + fsign(gearIndex)
        break
      end
      sharedFunctions.selectShiftPoints(gearIndex, isSportMode)
    end

    --shift up?
    if tmpEngineAV >= shiftBehavior.shiftUpAV and M.brake <= 0 and shiftPreventionData.wheelSlipShiftUp and abs(M.throttle - smoothedThrottle) < smoothThrottleUpShiftThreshold and gearIndex < gearbox.maxGearIndex and gearIndex > gearbox.minGearIndex then
      gearIndex = gearIndex + fsign(gearIndex)
      tmpEngineAV = relEngineAV * (gearbox.gearRatios[gearIndex] or 0)
      if tmpEngineAV < engine.idleAV then
        gearIndex = gearIndex - fsign(gearIndex)
      end
      sharedFunctions.selectShiftPoints(gearIndex, isSportMode)
    end
  end

  --enforce things like L and M modes
  gearIndex = min(max(gearIndex, automaticHandling.minGearIndex), automaticHandling.maxGearIndex)

  if gearbox[dct.primaryAccess.gearIndexName] ~= gearIndex then
    newDesiredGearIndex = gearIndex
    previousGearIndex = gearbox[dct.primaryAccess.gearIndexName]
    timer.shiftDelayTimer = 0
    updateGearbox = gearboxHandling.logic.whileShifting
  end

  -- Control clutch to buildup engine RPM
  local dctClutchRatio = 0
  if abs(gearIndex) == 1 and M.throttle > 0 then
    local ratio = max((engine.outputAV1 - clutchHandling.clutchLaunchStartAV * (1 + M.throttle)) / (clutchHandling.clutchLaunchTargetAV * (1 + clutchHandling.clutchLaunchIFactor)), 0)
    clutchHandling.clutchLaunchIFactor = min(clutchHandling.clutchLaunchIFactor + dt * 0.5, 1)
    dctClutchRatio = math.min(math.max(ratio * ratio, 0), 1)
  else
    clutchHandling.clutchLaunchIFactor = 0
    dctClutchRatio = 1
  end

  --always prevent stalling
  local stallPrevent = min(max((engine.outputAV1 * 0.95 - engine.idleAV) / (engine.idleAV * 0.05), 0), 1)
  dctClutchRatio = min(dctClutchRatio, stallPrevent * stallPrevent)

  electrics.values[dct.primaryAccess.clutchRatioName] = dctClutchRatio
  electrics.values[dct.secondaryAccess.clutchRatioName] = 0
  M.clutchRatio = 1
  currentGearIndex = (automaticHandling.mode == "N" or automaticHandling.mode == "P") and 0 or gearIndex
  gearbox.gearIndex = gearIndex --just so that the DCT can always present the "active" gear to the outside world
  gearbox.gearRatio = gearbox.gearRatios[gearIndex]

  dctFunctions.dctPredictNextGear()
end

dctFunctions.updateDCTWhileShifting = function(dt)
  M.throttle = inputValues.throttle
  -- secondary clutch closes while primary opens -> in gear update once fully closed
  local primaryGearIndex = gearbox[dct.primaryAccess.gearIndexName]
  local secondaryGearIndex = gearbox[dct.secondaryAccess.gearIndexName]

  if newDesiredGearIndex ~= secondaryGearIndex and timer.shiftDelayTimer <= 0 then
    --find out if our desired gear is actually on the secondary shaft
    local sameShaft = newDesiredGearIndex % 2 == secondaryGearIndex % 2
    --if so, we can directly shift to that desired gear on the secondary shaft
    --if not, we need to use a helper gear first which makes the actually desired gear part of the secondary shaft
    local newGearIndex = sameShaft and newDesiredGearIndex or primaryGearIndex + fsign(newDesiredGearIndex - primaryGearIndex)
    if secondaryGearIndex ~= newGearIndex then
      gearbox[dct.secondaryAccess.setGearIndexName](gearbox, newGearIndex)
      timer.shiftDelayTimer = timerConstants.shiftDelay
    end
  end

  local canShift = true
  local secondaryGearRatio = gearbox.gearRatios[secondaryGearIndex]
  local targetAV = secondaryGearRatio * gearbox.outputAV1
  if targetAV > engine.outputAV1 and not clutchHandling.didRevMatch then
    M.throttle = clutchHandling.revMatchThrottle
    electrics.values[dct.primaryAccess.clutchRatioName] = 0
    canShift = engine.outputAV1 >= targetAV or targetAV > engine.maxAV
    clutchHandling.didRevMatch = canShift
  end

  if timer.shiftDelayTimer <= 0 and canShift then
    local clutchRatio = min(electrics.values[dct.secondaryAccess.clutchRatioName] + (1 / dct.clutchTime) * dt, 1)
    local stallPrevent = min(max((engine.outputAV1 * 0.9 - engine.idleAV) / (engine.idleAV * 0.1), 0), 1)
    electrics.values[dct.primaryAccess.clutchRatioName] = min(1 - clutchRatio, stallPrevent * stallPrevent)
    electrics.values[dct.secondaryAccess.clutchRatioName] = min(clutchRatio, stallPrevent * stallPrevent)
    if clutchRatio == 1 or stallPrevent < 1 then
      dct.primaryAccess, dct.secondaryAccess = dct.secondaryAccess, dct.primaryAccess
      primaryGearIndex = gearbox[dct.primaryAccess.gearIndexName]
      clutchHandling.didRevMatch = false

      if newDesiredGearIndex == primaryGearIndex then
        updateGearbox = gearboxHandling.logic.inGear
        timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
      end
    end
  end
  gearbox.gearIndex = primaryGearIndex --just so that the DCT can always present the "active" gear to the outside world
end

--------------------------------------
-----------------CVT------------------
--------------------------------------

local cvtFunctions = {}

cvtFunctions.applyCVTGearboxMode = function()
  local autoIndex = automaticHandling.modeIndexLookup[automaticHandling.mode]
  if autoIndex then
    automaticHandling.modeIndex = min(max(autoIndex, 1), #automaticHandling.modes)
    automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]
  end

  if automaticHandling.mode == "P" then
    gearbox:setMode("park")
  elseif automaticHandling.mode == "N" then
    gearbox:setMode("neutral")
  elseif automaticHandling.mode == "R" then
    gearbox:setMode("reverse")
  else
    gearbox:setMode("drive")
  end
end

cvtFunctions.shiftUpCVT = function()
  if automaticHandling.mode == "N" then
    timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
  end

  automaticHandling.modeIndex = min(automaticHandling.modeIndex + 1, #automaticHandling.modes)
  automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]

  cvtFunctions.applyCVTGearboxMode()
  sharedFunctions.applyGearboxModeRestrictions()
end

cvtFunctions.shiftDownCVT = function()
  if automaticHandling.mode == "N" then
    timer.gearChangeDelayTimer = timerConstants.gearChangeDelay
  end

  automaticHandling.modeIndex = max(automaticHandling.modeIndex - 1, 1)
  automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]

  cvtFunctions.applyCVTGearboxMode()
  sharedFunctions.applyGearboxModeRestrictions()
end

cvtFunctions.shiftToGearIndexCVT = function(index)
  local desiredMode = automaticHandling.hShifterModeLookup[index]
  if not desiredMode or not automaticHandling.existingModeLookup[desiredMode] then
    if desiredMode and not automaticHandling.existingModeLookup[desiredMode] then
      gui.message({txt = "vehicle.drivetrain.cannotShiftAuto", context = {mode = desiredMode}})
    end
    desiredMode = "N"
  end
  automaticHandling.mode = desiredMode

  cvtFunctions.applyCVTGearboxMode()
  sharedFunctions.applyGearboxModeRestrictions()
end

cvtFunctions.updateCVTInGearArcade = function(dt)
  M.throttle = inputValues.throttle
  local gearIndex = automaticHandling.cvtGearIndexLookup[automaticHandling.mode]
  -- driving backwards? - only with automatic shift - for obvious reasons ;)
  if (gearIndex < 0 and smoothedAvgAV <= 0.15) or (gearIndex <= 0 and smoothedAvgAV < -1) then
    M.throttle, M.brake = M.brake, M.throttle
  end

  --interpolate based on throttle between high/low ranges
  local throttleCubed = M.throttle * M.throttle * M.throttle
  local targetAV = cvtHandling.lowAV + (cvtHandling.highAV - cvtHandling.lowAV) * throttleCubed
  local targetAVSmooth = smoother.cvtTargetAVSmoother:get(targetAV,dt)
  local engineAV = engine.outputAV1
  local avError = ((targetAVSmooth - engineAV)) * cvtHandling.aggression

  gearbox:setGearRatio(smoother.cvtGearRatioSmoother:get(gearbox.gearRatio + avError * dt,dt))

  if torqueConverterHandling.hasLockup and gearIndex >= torqueConverterHandling.lockupMinGear then
    electrics.values.lockupClutchRatio = min(max((engineAV - torqueConverterHandling.lockupAV) / torqueConverterHandling.lockupRange, 0), 1)
  else
    electrics.values.lockupClutchRatio = 0
  end

  -- neutral gear handling
  if abs(gearbox.gearIndex) <= 1 and timer.neutralSelectionDelayTimer <= 0 then
    if automaticHandling.mode ~= "P" and abs(smoothedAvgAV) < automaticHandling.arcadeAutoBrakeAVThreshold and M.throttle <= 0 then
      M.brake = max(M.brake, automaticHandling.arcadeAutoBrakeAmount)
      gearIndex = 0
      automaticHandling.mode = "N"
      cvtFunctions.applyCVTGearboxMode()
    end

    if inputValues.throttle > 0 and smoothedAvgAV > -1 and gearIndex < 1 then
      gearIndex = 1
      timer.neutralSelectionDelayTimer = timerConstants.neutralSelectionDelay
      automaticHandling.mode = "D"
      cvtFunctions.applyCVTGearboxMode()
    end

    if input.brake > 0 and inputValues.throttle <= 0 and smoothedAvgAV <= 0.15 and gearIndex > -1 then
      gearIndex = -1
      timer.neutralSelectionDelayTimer = timerConstants.neutralSelectionDelay
      automaticHandling.mode = "R"
      cvtFunctions.applyCVTGearboxMode()
    end
  end

  currentGearIndex = (automaticHandling.mode == "N" or automaticHandling.mode == "P") and 0 or gearIndex
end

cvtFunctions.updateCVTInGear = function(dt)
  M.throttle = inputValues.throttle
  local gearIndex = automaticHandling.cvtGearIndexLookup[automaticHandling.mode]
  --interpolate based on throttle between high/low ranges
  local throttleCubed = automaticHandling.mode ~= "S" and (M.throttle * M.throttle * M.throttle) or 1
  local targetAV = cvtHandling.lowAV + (cvtHandling.highAV - cvtHandling.lowAV) * throttleCubed
  local targetAVSmooth = smoother.cvtTargetAVSmoother:get(targetAV,dt)
  local engineAV = engine.outputAV1
  local avError = ((targetAVSmooth - engineAV)) * cvtHandling.aggression

  gearbox:setGearRatio(smoother.cvtGearRatioSmoother:get(gearbox.gearRatio + avError * dt, dt))

  if torqueConverterHandling.hasLockup and gearIndex >= torqueConverterHandling.lockupMinGear then
    electrics.values.lockupClutchRatio = min(max((engineAV - torqueConverterHandling.lockupAV) / torqueConverterHandling.lockupRange, 0), 1)
  else
    electrics.values.lockupClutchRatio = 0
  end

  currentGearIndex = (automaticHandling.mode == "N" or automaticHandling.mode == "P") and 0 or gearIndex
end

--------------------------------------
--------------------------------------
--------------------------------------

local function getUIGearName()
  if gearbox.type == "manualGearbox" then
    return gearbox.gearIndex
  else
    local modePrefix = ""
    if automaticHandling.mode == "S" then
      modePrefix = "S"
    elseif type(automaticHandling.mode) == "number" then
      modePrefix = "M"
    end
    if gearbox.type == "automaticGearbox"then
      return modePrefix ~= "" and modePrefix..tostring(gearbox.gearIndex) or automaticHandling.mode
    elseif gearbox.type == "dctGearbox" then
      return modePrefix ~= "" and modePrefix..tostring(gearbox[dct.primaryAccess.gearIndexName]) or automaticHandling.mode
    elseif gearbox.type == "cvtGearbox" then
      return automaticHandling.mode
    end
  end
end

local function handleStalling(dt)
  timer.stalledEngineMessageTimer = max(timer.stalledEngineMessageTimer - dt, 0)
  if engine.isStalled and not engine.isDisabled then
    if gearboxHandling.behavior == "arcade" and timer.stalledEngineMessageTimer <= 0 then
      if not isStartingEngine and not engineManuallyTurnedOff then
        --gearbox:setGearIndex(0)
        engine:activateStarter()
        isStartingEngine = true
        gui.message({txt="vehicle.drivetrain.stalledStarting"}, 2, "stalling")
        timer.stalledEngineMessageTimer = 1.8
      end
    elseif timer.stalledEngineMessageTimer <= 0 then
      local message
      if gearboxHandling.autoClutch then
        message = "vehicle.drivetrain.stalledAutoClutch"
      else
        message = "vehicle.drivetrain.stalled"
      end
      gui.message({txt = message}, 2, "stalling")
      timer.stalledEngineMessageTimer = 1.8
    end
  elseif isStartingEngine then
    engine:deactivateStarter()
    isStartingEngine = false
  end

  if engineManuallyTurnedOff and engineWasStalled and not engine.isStalled then
    engineManuallyTurnedOff = false
  end
  engineWasStalled = engine.isStalled
end

local function updateWheelSlip(dt)
  local overallWheelSlip = 0
  for _,wi in pairs(wheels.wheels) do
    overallWheelSlip = overallWheelSlip + wi.slipEnergy
  end

  shiftPreventionData.wheelSlipShiftDown = true
  --  if overallWheelSlip > shiftPreventionData.wheelSlipDownThreshold then
  --    if abs(gearbox.outputAV1) < 10 then
  --      shiftPreventionData.wheelSlipShiftDown = false
  --    end
  --  end

  overallWheelSlip = smoother.wheelSlipShiftUpSmoother:get(overallWheelSlip, dt)

  shiftPreventionData.wheelSlipShiftUp = overallWheelSlip < shiftPreventionData.wheelSlipUpThreshold
end

local function updateGFX(dt)
  inputValues.throttle = electrics.values.throttleOverride or min(max(input.throttle or 0, 0), 1)
  M.brake = min(max(input.brake or 0, 0), 1)
  inputValues.clutch = min(max(input.clutch or 0, 0), 1)

  smoothedAvgAV = smoother.avgAVsmoother:get(gearbox.outputAV1, dt)

  timer.gearChangeDelayTimer = max(timer.gearChangeDelayTimer - dt, 0)
  timer.shiftDelayTimer = max(timer.shiftDelayTimer - dt, 0)
  timer.neutralSelectionDelayTimer = max(timer.neutralSelectionDelayTimer - dt, 0)

  updateWheelSlip(dt)
  updateGearbox(dt)
  handleStalling(dt)

  smoothedThrottle = smoother.throttleSmoother:getUncapped(M.throttle, dt)

  M.fireEngineTemperature = engine.thermals and engine.thermals.exhaustTemperature or obj:getEnvTemperature()

  if isFrozen then
    M.brake = 1
  end

  energyStorageData.ratio = 0
  energyStorageData.volume = 0

  for _,s in pairs(engine.energyStorage or {}) do
    local energyStorage = energyStorage.getStorage(s)
    if energyStorage then
      energyStorageData.ratio = energyStorageData.ratio + energyStorage.remainingRatio
      energyStorageData.volume = energyStorageData.volume + energyStorage.remainingVolume
    end
  end
  energyStorageData.ratio = energyStorageData.ratio * energyStorageData.invEnergyStorageCount

  electrics.values.fuel = energyStorageData.ratio or 0
  electrics.values.lowfuel = electrics.values.fuel < 0.1
  electrics.values.fuelCapacity = energyStorageData.invEnergyStorageCount > 0 and (energyStorageData.capacity or 0) or 1
  electrics.values.fuelVolume = energyStorageData.volume

  electrics.values.throttle = M.throttle
  electrics.values.brake = M.brake
  electrics.values.clutch = 1 - M.clutchRatio
  electrics.values.clutchRatio = M.clutchRatio
  electrics.values.gear = getUIGearName()
  electrics.values.gear_A = (automaticHandling.modeIndex - 1) / (#automaticHandling.modes - 1)
  electrics.values.gearIndex = currentGearIndex
  electrics.values.rpm = engine.outputAV1 * constants.avToRPM
  electrics.values.oiltemp = engine.thermals and engine.thermals.oilTemperature or 0
  electrics.values.watertemp = engine.thermals and engine.thermals.coolantTemperature or 0
  electrics.values.checkengine = engine.isDisabled
  electrics.values.ignition = not engine.isDisabled
  electrics.values.engineThrottle = engine.isDisabled and 0 or M.throttle
  electrics.values.engineLoad = engine.isDisabled and 0 or engine.engineLoad
  electrics.values.running = electrics.values.ignition
  electrics.values.radiatorFanSpin = engine.thermals and engine.thermals.radiatorFanSpin or 0

  M.engineInfo = {
    engine.idleAV * constants.avToRPM
    , engine.maxAV * constants.avToRPM
    , 0--v.data.engine.shiftUpRPM
    , 0--v.data.engine.shiftDownRPM
    , electrics.values.rpm
    , electrics.values.gear
    , gearbox.maxGearIndex
    , abs(gearbox.minGearIndex)
    , engine.outputTorque1
    , gearbox.outputTorque1
    , obj:getVelocity():length()  -- airspeed
    , electrics.values.fuelVolume
    , electrics.values.fuelCapacity
    , "manual" --v.data.engine.transmissionType
    , obj:getID()
    , drivetrain.brakeSpecificFuelConsumption
    , electrics.values.gearIndex
    , engine.isStalled and 0 or 1
    , electrics.values.engineLoad
  }

  --if playWheelShiftLEDs and playerInfo.anyPlayerSeated then
  -- disabled until logitech drivers bug is fixed allowing to not mess up the steering angles and the ffb
  -- obj:playRPMLeds(engine.outputAV1 or 0, (engine.maxAV or 1) * 0.8, engine.maxAV or 1) -- currentValue, firstLEDValue, lastLEDValue)
  --end

--  if streams.gearboxData > 0 then
--    gui.send('gearboxData', {
--        gearIndex = gearbox.gearIndex or "",
--        clutchRatio = M.clutchRatio or "",
--        automaticMode = automaticHandling.mode or "",
--        dctGearIndex1 = gearbox.gearIndex1 or "",
--        dctGearIndex2 = gearbox.gearIndex2 or "",
--        dctClutchRatio1 = gearbox.clutchRatio1 or "",
--        dctClutchRatio2 = gearbox.clutchRatio2 or "",
--      })
--  end
end

local function updateGFXNoGearbox()
  inputValues.throttle = min(max(input.throttle, 0), 1)
  M.brake = min(max(input.brake, 0), 1)
  inputValues.clutch = min(max(input.clutch, 0), 1)
  M.throttle = inputValues.throttle

  M.fireEngineTemperature = engine.thermals.exhaustTemperature

  if isFrozen then
    M.brake = 1
  end

  electrics.values.throttle = M.throttle
  electrics.values.brake = M.brake
  electrics.values.clutch = 1 - M.clutchRatio
  electrics.values.clutchRatio = M.clutchRatio
  electrics.values.gear = "N"
  electrics.values.gearIndex = 0
  electrics.values.rpm = engine.outputAV1 * constants.avToRPM
  electrics.values.oiltemp = engine.thermals.oilTemperature
  electrics.values.watertemp = engine.thermals.coolantTemperature
  electrics.values.radiatorFanSpin = 0

  M.engineInfo = {
    engine.idleAV * constants.avToRPM
    , engine.maxAV * constants.avToRPM
    , 0
    , 0
    , electrics.values.rpm
    , electrics.values.gear
    , 0
    , 0
    , engine.outputTorque1
    , 0
    , obj:getVelocity():length()  -- airspeed
    , 0
    , 0
    --, v.data.engine.transmissionType
    , "manual"
    , obj:getID()
    , drivetrain.brakeSpecificFuelConsumption
    , 0
    , 1
    , 0
  }
end

local function updateGFXNoEngine(dt)
  inputValues.throttle = 0
  inputValues.clutch = 0
  M.throttle = inputValues.throttle
  M.brake = 0
  M.fireEngineTemperature = 0

  if isFrozen then
    M.brake = 1
  end

  electrics.values.throttle = M.throttle
  electrics.values.brake = M.brake
  electrics.values.clutch = 1 - M.clutchRatio
  electrics.values.clutchRatio = M.clutchRatio
  electrics.values.gear = "N"
  electrics.values.gearIndex = 0
  electrics.values.rpm = 0
  electrics.values.oiltemp = 0
  electrics.values.watertemp = 0
  electrics.values.radiatorFanSpin = 0

  M.engineInfo = {
    0
    , 0
    , 0
    , 0
    , 0
    , electrics.values.gear
    , 0
    , 0
    , 0
    , 0
    , obj:getVelocity():length()  -- airspeed
    , 0
    , 0
    --, v.data.engine.transmissionType
    , "manual"
    , obj:getID()
    , 0
    , 0
    , 1
    , 0
  }
end

local function buildGearBoxUpdates()
  if not gearboxHandling.availableLogic[gearbox.type] then
    log("E", "carController.buildGearBoxUpdates", "Unknown gearbox type: "..gearbox.type)
    return
  end
  if not gearboxHandling.availableLogic[gearbox.type][gearboxHandling.behavior] then
    log("E", "carController.buildGearBoxUpdates", "Unknown gearbox behavior: "..(gearboxHandling.behavior or "nil"))
    return
  end

  gearboxHandling.logic = gearboxHandling.availableLogic[gearbox.type][gearboxHandling.behavior]
  gearboxHandling.logic.gearboxBehaviorChanged()
end

local function shiftUp()
  if not gearbox then
    return
  end

  gearboxHandling.logic.shiftUp()
end

local function shiftDown()
  if not gearbox then
    return
  end

  gearboxHandling.logic.shiftDown()
end

local function shiftToGearIndex(index)
  if not gearbox then
    return
  end

  gearboxHandling.logic.shiftToGearIndex(index)
end

local function getGearboxBehaviorName()
  if gearboxHandling.behavior == "arcade" then
    return "vehicle.drivetrain.shifterModeNameArcade"
  elseif gearboxHandling.behavior == "realistic" then
    return "vehicle.drivetrain.shifterModeNameRealistic"
  end

  return "vehicle.drivetrain.shifterModeNameUnknown"
end

local function setGearboxBehavior(behavior)
  if not gearbox then
    return
  end

  if not gearboxHandling.behaviorLookup[behavior] then
    log("E", "vehicleController.setGearboxBehavior", "Unknown gearbox behavior: "..(behavior or "nil"))
    return
  end

  gearboxHandling.behavior = behavior
  gui.message({txt="vehicle.drivetrain.shifterModeChanged", context = { shifterModeName = getGearboxBehaviorName()}}, 5, "shiftermode")
  gearboxHandling.previousBehavior = gearboxHandling.behavior

  buildGearBoxUpdates()
  updateGearbox = gearboxHandling.logic.inGear
end

local function cycleGearboxBehaviors()
  if not gearbox then
    return
  end

  local found = false
  local newBehavior = gearboxHandling.behavior
  for _,v in pairs(gearboxHandling.behaviors) do
    if found then
      newBehavior = v
      found = false
      break
    elseif gearboxHandling.behavior == v then
      found = true
    end
  end

  if found then
    newBehavior = gearboxHandling.behaviors[next(gearboxHandling.behaviors)]
  end

  setGearboxBehavior(newBehavior)
end

local function switchToRealisticBehavior(gearIndex)
  gui.message({txt = "vehicle.drivetrain.usingHshifter"})
  setGearboxBehavior("realistic")
  gearboxHandling.logic.shiftToGearIndex(gearIndex)
end

local function warnCannotShiftSequential()
  gui.message({txt = "vehicle.drivetrain.cannotShiftSequential", context = {shifterModeName = getGearboxBehaviorName()}})
end

local function setStarter(enabled)
  if not engine then
    return
  end

  if enabled then
    if not engine.isStalled then
      engineManuallyTurnedOff = true
    end
    engine:activateStarter()
  else
    engine:deactivateStarter()
  end
end

local function setFreeze(mode)
  isFrozen = mode == 1
  if gearbox then
    gearbox:setLock(isFrozen)
  end
end

local function setEngineIgnition(value)
  engineManuallyTurnedOff = not value
  engine:setIgnition(value and 1 or 0)
end

local function sendTorqueData()
  if not playerInfo.firstPlayerSeated then return end
  if engine then
    engine:sendTorqueData()
  end
end

local function settingsChanged(noRefresh)
  if not noRefresh then
    settings.refresh()
  end
  gearboxHandling.autoClutch = settings.getValue("autoClutch", true)
  gearboxHandling.autoThrottle = settings.getValue("autoThrottle", true)
  gearboxHandling.autoShiftPrevention = settings.getValue("autoShiftPrevention", true)
  --playWheelShiftLEDs = settings.getValue('rpmLedsEnabled', false)
end

local function calculateOptimalLoadShiftPoints()
  local torqueCurve = engine.torqueData.curves[engine.torqueData.finalCurveName].torque
  for k,v in pairs(gearbox.gearRatios) do
    local shiftUp = nil
    local shiftDown = nil
    if v ~= 0 then
      local currentGearRatio = v
      local nextGearRatio = gearbox.gearRatios[k + 1] or 0
      local previousGearRatio = gearbox.gearRatios[k - 1] or 0

      for i = 100, engine.maxRPM - 100, 50 do
        local currentWheelTorque = torqueCurve[i] * currentGearRatio
        local nextGearRPM = min(max(math.floor(i * ( nextGearRatio / currentGearRatio)),1), engine.maxRPM)
        local previousGearRPM = min(max(math.floor(i * (previousGearRatio / currentGearRatio)),1), engine.maxRPM)
        local nextWheelTorque = torqueCurve[nextGearRPM] * nextGearRatio
        local previousWheelTorque = torqueCurve[previousGearRPM] * previousGearRatio

        if currentWheelTorque < nextWheelTorque and not shiftUp and currentWheelTorque > 0 and currentGearRatio * nextGearRatio > 0 then
          shiftUp = i
        end

        if previousWheelTorque > currentWheelTorque * 1.05 and previousGearRPM < engine.maxRPM * 0.95 and currentWheelTorque > 0 and currentGearRatio * previousGearRatio > 0 then
          shiftDown = i
        end
      end
    end

    shiftPoints[k].highShiftUpAV = (shiftUp or engine.maxRPM * 0.98) * constants.rpmToAV
    shiftPoints[k].highShiftDownAV = (shiftDown or 0) * constants.rpmToAV

    --print(string.format("Gear %d: Up: %d, Down: %d",k,shiftPoints[k].highShiftUpAV * constants.avToRPM,shiftPoints[k].highShiftDownAV * constants.avToRPM))
  end
end

local function init(jbeamData)
  gearboxHandling.availableLogic =
  {
    manualGearbox =
    {
      arcade =
      {
        inGear = manualFunctions.updateManualInGearArcade,
        whileShifting = manualFunctions.updateManualWhileShiftingArcade,
        shiftUp = warnCannotShiftSequential,
        shiftDown = warnCannotShiftSequential,
        shiftToGearIndex = switchToRealisticBehavior,
        gearboxBehaviorChanged = manualFunctions.manualGearboxBehaviorChanged
      },
      realistic =
      {
        inGear = manualFunctions.updateManualInGear,
        whileShifting = manualFunctions.updateManualWhileShifting,
        shiftUp = manualFunctions.shiftUpManual,
        shiftDown = manualFunctions.shiftDownManual,
        shiftToGearIndex = manualFunctions.shiftToGearIndexManual,
        gearboxBehaviorChanged = manualFunctions.manualGearboxBehaviorChanged
      }
    },
    automaticGearbox =
    {
      arcade =
      {
        inGear = automaticFunctions.updateAutomaticInGearArcade,
        whileShifting = automaticFunctions.updateAutomaticWhileShiftingArcade,
        shiftUp = warnCannotShiftSequential,
        shiftDown = warnCannotShiftSequential,
        shiftToGearIndex = switchToRealisticBehavior,
        gearboxBehaviorChanged = nop
      },
      realistic =
      {
        inGear = automaticFunctions.updateAutomaticInGear,
        whileShifting = automaticFunctions.updateAutomaticWhileShifting,
        shiftUp = automaticFunctions.shiftUpAutomatic,
        shiftDown = automaticFunctions.shiftDownAutomatic,
        shiftToGearIndex = automaticFunctions.shiftToGearIndexAutomatic,
        gearboxBehaviorChanged = nop
      }
    },
    dctGearbox =
    {
      arcade =
      {
        inGear = dctFunctions.updateDCTInGearArcade,
        whileShifting = dctFunctions.updateDCTWhileShiftingArcade,
        shiftUp = warnCannotShiftSequential,
        shiftDown = warnCannotShiftSequential,
        shiftToGearIndex = switchToRealisticBehavior,
        gearboxBehaviorChanged = nop
      },
      realistic =
      {
        inGear = dctFunctions.updateDCTInGear,
        whileShifting = dctFunctions.updateDCTWhileShifting,
        shiftUp = dctFunctions.shiftUpDCT,
        shiftDown = dctFunctions.shiftDownDCT,
        shiftToGearIndex = dctFunctions.shiftToGearIndexDCT,
        gearboxBehaviorChanged = nop
      }
    },
    cvtGearbox =
    {
      arcade =
      {
        inGear = cvtFunctions.updateCVTInGearArcade,
        whileShifting = nop,
        shiftUp = warnCannotShiftSequential,
        shiftDown = warnCannotShiftSequential,
        shiftToGearIndex = switchToRealisticBehavior,
        gearboxBehaviorChanged = nop
      },
      realistic =
      {
        inGear = cvtFunctions.updateCVTInGear,
        whileShifting = nop,
        shiftUp = cvtFunctions.shiftUpCVT,
        shiftDown = cvtFunctions.shiftDownCVT,
        shiftToGearIndex = cvtFunctions.shiftToGearIndexCVT,
        gearboxBehaviorChanged = nop
      }
    },
  }

  M.throttle = 0
  M.brake = 0
  M.clutchRatio = 0

  timer.shiftDelayTimer = 0
  timer.gearChangeDelayTimer = 0
  timer.neutralSelectionDelayTimer = 0
  timer.stalledEngineMessageTimer = 0
  clutchHandling.preShiftClutchRatio = 0
  clutchHandling.shiftState = "clutchIn"
  clutchHandling.didRevMatch = false
  newDesiredGearIndex = 0
  previousGearIndex = 0
  engineManuallyTurnedOff = false
  engineWasStalled = false
  isStartingEngine = false
  isStartingEngine = false

  settingsChanged(true)

  smoothedThrottle = 0
  smoothedAvgAV = 0

  energyStorageData = {
    ratio = 0,
    volume = 0,
    capacity = 0,
    invEnergyStorageCount = 0,
  }

  local throttleSmoothingIn = jbeamData.gearboxDecisionSmoothingDown or 2
  local throttleSmoothingOut = jbeamData.gearboxDecisionSmoothingUp or 5
  local cvtGearRatioSmoothingIn = jbeamData.cvtGearRatioSmoothingIn or 20
  local cvtGearRatioSmoothingOut = jbeamData.cvtGearRatioSmoothingOut or 5
  local cvtTargetAVSmoothingIn = jbeamData.cvtTargetAVSmoothingIn or 0.5
  local cvtTargetAVSmoothingOut = jbeamData.cvtTargetAVSmoothingOut or 2
  local wheelSlipShiftUpSmoothingIn = jbeamData.wheelSlipShiftUpSmoothingIn or 2
  local wheelSlipShiftUpSmoothingOut = jbeamData.wheelSlipShiftUpSmoothingOut or 20

  smoother.throttleSmoother = newTemporalSmoothing(throttleSmoothingIn, throttleSmoothingOut)
  smoother.avgAVsmoother = newTemporalSmoothingNonLinear(5, 5)
  smoother.cvtGearRatioSmoother = newTemporalSmoothingNonLinear(cvtGearRatioSmoothingIn, cvtGearRatioSmoothingOut)
  smoother.cvtTargetAVSmoother = newTemporalSmoothingNonLinear(cvtTargetAVSmoothingIn, cvtTargetAVSmoothingOut)
  smoother.wheelSlipShiftUpSmoother = newTemporalSmoothingNonLinear(wheelSlipShiftUpSmoothingIn, wheelSlipShiftUpSmoothingOut)

  shiftPreventionData.wheelSlipUpThreshold = jbeamData.wheelSlipUpThreshold or 150
  shiftPreventionData.wheelSlipDownThreshold = jbeamData.wheelSlipDownThreshold or 20

  engine = powertrain.getDevice(expectedDeviceNames.engine)
  gearbox = powertrain.getDevice(expectedDeviceNames.gearbox)
  local torqueConverter = powertrain.getDevice(expectedDeviceNames.torqueConverter)

  local hasEngine = engine ~= nil
  local hasGearbox = gearbox ~= nil
  local hasTorqueConverter = torqueConverter ~= nil

  if hasGearbox then
    gearboxType = gearbox.type

    local shiftRPMNames = {"lowShiftDownRPM", "highShiftDownRPM", "lowShiftUpRPM", "highShiftUpRPM"}
    local defaultShiftPoints = {lowShiftDownRPM = 2000, highShiftDownRPM = 3500, lowShiftUpRPM = 2500, highShiftUpRPM = 5000}
    local gearCount = gearbox.gearCount + 1
    for _,v in pairs(shiftRPMNames) do
      if type(jbeamData[v]) ~= "table" then
        local shiftRPM = jbeamData[v] or defaultShiftPoints[v]
        jbeamData[v] = {}
        for _ = 0, gearCount, 1 do
          table.insert(jbeamData[v], shiftRPM)
        end
      else
        if tableSize(jbeamData[v]) ~= gearCount then
          for i = 1, gearCount, 1 do
            if not jbeamData[v][i] then
              jbeamData[v][i] = jbeamData[v][i-1] or 0
            end
          end
        end
      end
    end

    local gearCounter = 1
    for i = gearbox.minGearIndex, gearbox.maxGearIndex, 1 do
      shiftPoints[i] =
      {
        lowShiftDownAV = jbeamData.lowShiftDownRPM[gearCounter] * constants.rpmToAV,
        highShiftDownAV = (jbeamData.highShiftDownRPM[gearCounter] or 0) * constants.rpmToAV,
        lowShiftUpAV = jbeamData.lowShiftUpRPM[gearCounter] * constants.rpmToAV,
        highShiftUpAV = (jbeamData.highShiftUpRPM[gearCounter] or 0) * constants.rpmToAV
      }
      gearCounter = gearCounter + 1
    end

    --dump(jbeamData)

    if jbeamData.calculateOptimalLoadShiftPoints then
      calculateOptimalLoadShiftPoints()
    end

    timerConstants.shiftDelay = jbeamData.transmissionShiftDelay or 0.22
    timerConstants.gearChangeDelay = jbeamData.transmissionGearChangeDelay or 1
    timerConstants.neutralSelectionDelay = jbeamData.neutralSelectionDelay or 0.5

    if gearboxType ~= "manualGearbox" then
      automaticHandling.availableModeLookup = {}
      for _,v in pairs(automaticHandling.availableModes) do
        automaticHandling.availableModeLookup[v] = true
      end

      automaticHandling.modes = {}
      automaticHandling.modeIndexLookup = {}
      local modes = jbeamData.automaticModes or "PRNDS21M"
      local modeCount = #modes
      local modeOffset = 0
      for i = 1, modeCount do
        local mode = modes:sub(i,i)
        if automaticHandling.availableModeLookup[mode] then
          if mode ~= "M" then
            automaticHandling.modes[i + modeOffset] = mode
            automaticHandling.modeIndexLookup[mode] = i + modeOffset
            automaticHandling.existingModeLookup[mode] = true
          else
            for j = 1, gearbox.maxGearIndex, 1 do
              local manualMode = "M"..tostring(j)
              local manualModeIndex = i + j - 1
              automaticHandling.modes[manualModeIndex] = manualMode
              automaticHandling.modeIndexLookup[manualMode] = manualModeIndex
              automaticHandling.existingModeLookup[manualMode] = true
              modeOffset = j - 1
            end
          end
        else
          print("unknown auto mode: "..mode)
        end
      end

      if hasTorqueConverter then
        torqueConverterHandling.lockupAV = (jbeamData.torqueConverterLockupRPM or 0) * constants.rpmToAV
        torqueConverterHandling.lockupRange = (jbeamData.torqueConverterLockupRange or (torqueConverterHandling.lockupAV * 0.2 * constants.avToRPM)) * constants.rpmToAV
        torqueConverterHandling.lockupMinGear = jbeamData.torqueConverterLockupMinGear or 0
        torqueConverterHandling.hasLockup = torqueConverterHandling.lockupAV > 0
      end

      local defaultMode = jbeamData.defaultAutomaticMode or "N"
      automaticHandling.modeIndex = string.find(modes, defaultMode)
      automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]
      automaticHandling.maxGearIndex = gearbox.maxGearIndex
      automaticHandling.minGearIndex = gearbox.minGearIndex
      if gearboxType == "automaticGearbox" then
        automaticFunctions.applyAutoGearboxMode()
      elseif gearboxType == "dctGearbox" then
        dct.clutchTime = jbeamData.dctClutchTime or 0.1

        dct.primaryAccess = dct.access1
        dct.secondaryAccess = dct.access2
        dctFunctions.applyDCTGearboxMode()
      elseif gearboxType == "cvtGearbox" then
        cvtHandling.aggression = jbeamData.cvtAggression or 0.5
        cvtHandling.highAV = (jbeamData.cvtHighRPM or (shiftPoints[1] and shiftPoints[1].highShiftUpAV / constants.rpmToAV) or 0) * constants.rpmToAV
        cvtHandling.lowAV = (jbeamData.cvtLowRPM or (shiftPoints[1] and shiftPoints[1].lowShiftUpAV / constants.rpmToAV) or 0) * constants.rpmToAV

        smoother.cvtGearRatioSmoother:set(gearbox.maxGearRatio)
        smoother.cvtTargetAVSmoother:set(cvtHandling.lowAV)
        cvtFunctions.applyCVTGearboxMode()
      end
    end

    gearboxHandling.behaviorLookup = {}
    for  _,v in pairs(gearboxHandling.behaviors) do
      gearboxHandling.behaviorLookup[v] = true
    end
    setGearboxBehavior(gearboxHandling.previousBehavior or settings.getValue("defaultGearboxBehavior") or "arcade")
  end

  if hasEngine then
    clutchHandling.clutchLaunchTargetAV = (jbeamData.clutchLaunchTargetRPM or 3000) * constants.rpmToAV * 0.5
    clutchHandling.clutchLaunchStartAV = ((jbeamData.clutchLaunchStartRPM or 2000) * constants.rpmToAV - engine.idleAV) * 0.5
    clutchHandling.clutchLaunchIFactor = 0

    clutchHandling.clutchInRate = jbeamData.clutchInRate or 10
    clutchHandling.clutchOutRate = jbeamData.clutchOutRate or 10

    clutchHandling.revMatchThrottle = jbeamData.revMatchThrottle or 0.5

    local energyStorageCount = 0
    for _,s in pairs(engine.energyStorage or {}) do
      local energyStorage = energyStorage.getStorage(s)
      if energyStorage then
        energyStorageData.capacity = energyStorageData.capacity + energyStorage.capacity
        energyStorageCount = energyStorageCount + 1
      end
    end

    energyStorageData.invEnergyStorageCount = energyStorageCount > 0 and 1 / energyStorageCount or 0
  end

  sendTorqueData()

  if hasEngine and hasGearbox then
    M.updateGFX = updateGFX
  elseif hasEngine then
    M.updateGFX = updateGFXNoGearbox
  else
    M.updateGFX = updateGFXNoEngine
  end
end

local function onDeserialize(data)
  if data.previousGearboxBehavior then
    setGearboxBehavior(data.previousGearboxBehavior)
  end
end

local function onSerialize()
  return { previousGearboxBehavior = gearboxHandling.behavior }
end

M.init = init
M.updateGFX = nop
M.settingsChanged = settingsChanged

M.onDeserialize  = onDeserialize
M.onSerialize  = onSerialize

M.setEngineIgnition = setEngineIgnition

--Mandatory main controller API
M.shiftUp = shiftUp
M.shiftDown = shiftDown
M.shiftToGearIndex = shiftToGearIndex
M.cycleGearboxModes = cycleGearboxBehaviors
M.setGearboxMode = setGearboxBehavior
M.setStarter = setStarter
M.setFreeze = setFreeze
M.sendTorqueData = sendTorqueData
-------------------------------

return M
