-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

package.path = "lua/vehicle/?.lua;lua/vehicle/jbeam/?.lua;lua/common/?.lua;lua/common/socket/?.lua;lua/?.lua;?.lua"
package.cpath = ''
require('compatibility')

log = function(...) Lua:log(...) end
print = function(...) Lua:log('A', "print", tostring(...)) end

require('utils')
require('mathlib')

math.randomseed(os.time())

-- improve stacktraces
local STP = require "StackTracePlus"
debug.traceback = STP.stacktrace

perf = require("perf")
settings = require('simplesettings')
backwardsCompatibility = require("backwardsCompatibility")
objectId = obj:getID()

playerInfo = {
  seatedPlayers = {}, -- list of players seated in this vehicle; players are indexed from 0 to N (e.g. { [1]=true, [4]=true } for 2nd and 5th players)
  firstPlayerSeated = false,
  anyPlayerSeated = false,
}
lastDt = 1/20

loadingTimes = {} -- global intentionally, used all over the place

local initCalled = false
log_jbeam = nop -- intentionally global

-- step functions
-- Note: dts is simulation dt
function physicsStep(dts)
  powertrain.update(dts)
  wheels.update(dts)
  controller.update(dts)
  thrusters.update()
  hydros.update(dts)
  --commands.onPhysicsStep(dts)
end

-- This is called in the local scope, so it is NOT safe to do things that contact things outside the vehicle
-- Note: dt is simulation dt
function graphicsStep(dt)
  lastDt = dt
  sensors.updateGFX(dt) -- must be before input and ai
  mapmgr.sendTracking() -- must be before ai.updateGFX
  ai.updateGFX(dt) -- must be before input
  input.update(dt) -- must be as early as possible
  electrics.update(dt)
  controller.updateGFX(dt)
  extensions.hook('updateGFX', dt) -- must be before drivetrain and after electrics
  powertrain.updateGFX(dt)
  energyStorage.updateGFX(dt)
  drivetrain.updateGFX(dt)
  beamstate.updateGFX(dt) -- must be after drivetrain
  sounds.update(dt)
  hydros.updateGFX(dt) -- must be after (input, electrics) and before props
  thrusters.updateGFX() -- should be after extensions.hook
  if playerInfo.anyPlayerSeated then
    perf.update()
    gui.frameUpdated(obj:getRealdt())
  end

  --drivetrain.updateWheelsGFX(dt)
  wheels.updateGFX(dt)
  mapmgr.updateGFX()
  scenario.updateGFX()
  props.update()
  fire.updateGFX(dt)
  recovery.updateGFX(dt)
end

-- debug rendering
function debugDraw(x, y, z)
  local focusPos = float3(x,y,z)
  bdebug.debugDraw(focusPos)
  ai.debugDraw(focusPos)
  beamstate.debugDraw(focusPos)
  extensions.hook('onDebugDraw', focusPos)
end

function initSystems()
  backwardsCompatibility.init()

  material.init()
  wheels.init()
  powertrain.init()
  energyStorage.init()
  controller.init()

  wheels.initSecondStage()
  controller.initSecondStage()
  drivetrain.init()

  sensors.reset()
  beamstate.init()
  thrusters.init()
  hydros.init()
  input.init()
  sounds.init()
  props.init()
  electrics.init()
  recovery.init()
  bdebug.init()
  scenario.init()
  sensors.init()
  fire.init()
  gui.message("", 0, "^damage\\.") -- clear damage messages on vehicle restart
  extensions.hook('onInit')

  -- be sensitive about global writes from now on
  detectGlobalWrites()
  initCalled = true
end

function init(path, partConfigData)
  local hp1 = HighPerfTimer()
  local hp2 = HighPerfTimer()

  --perf.enable(1)

  if not obj then log('W', "default.init", "Error getting main object: unable to spawn") ; return end
  log('D', "default.init", "spawning vehicle "..tostring(path))

  -- we change the lookup path here, so it prefers the vehicle lua
  package.path = path .. "/lua/?.lua;".. package.path
  extensions = require("extensions")
  extensions.loadModulesInDirectory(path.."/lua")
  extensions.addModulePath("lua/vehicle/extensions/")
  extensions.loadModule('quickAccess')

  drivetrain = require("drivetrain")
  powertrain = require("powertrain")
  energyStorage = require("energyStorage")
  controller = require("controller")

  wheels     = require("wheels")
  sounds     = require("sounds")

  bdebug     = require("bdebug")
  input      = require("input")
  props      = require("props")

  particlefilter = require("particlefilter")
  particles  = require("particles")
  material   = require("material")
  v          = require("jbeam_main")
  electrics  = require("electrics")
  beamstate  = require("beamstate")
  sensors    = require("sensors")
  bullettime = require("bullettime") -- to be deprecated
  thrusters  = require("thrusters")
  hydros     = require("hydros")
  gui        = require("guihooks") -- do not change its name, the GUI callback will break otherwise
  partmgmt   = require("partmgmt") -- do not change its name, the GUI callback will break otherwise
  streams    = require("guistreams")
  guihooks   = gui -- legacy
  ai         = require("ai")
  recovery   = require("recovery")
  mapmgr     = require("mapmgr")
  scenario   = require("scenario")
  fire       = require("fire")
  commands   = require("commands")
  table.insert(loadingTimes, {'0 startup', hp1:stopAndReset()})

  -- care about the config before pushing to the physics
  if type(partConfigData) == 'string' and string.len(partConfigData) > 0 then
    local firstChar = string.sub(partConfigData,1,1)
    if firstChar == '{' or firstChar == '[' then
      log('D', "default.init", "  using partconfig data "..tostring(partConfigData)..' - path:'..tostring(path) )
      partmgmt.setConfig(unserialize(partConfigData), false)
    else
      log('D', "default.init", "  using partconfig filename "..tostring(partConfigData)..' - path:'..tostring(path) )
      partmgmt.load(partConfigData, false)
    end
  end

  -- experimental: vehicle file caching
  --[[
  local mp = require 'MessagePack'
  local vCacheFilename = 'vehicleCache.bin'
  if FS:fileExists(vCacheFilename) then
    local f = io.open(vCacheFilename, 'r')
    if f then
      v.vehicles = mp.unpack(f:read('*a'))
      f:close()
    end
  end
  --]]

  -- this filters the Debug messages out
  log_jbeam = function(level, source, msg)
    if level == 'D' then return end
    log(level, source, msg)
  end
  if settings and settings.getValue('creatorMode') == true then
    extensions.loadModule('creatorMode')
  end

  if settings and settings.getValue('externalUi') == true then
    extensions.loadModule('extUI')
  end

  --extensions.loadModule("api")

  --if v.vehicles == nil then
    -- load jbeam files
    if not v.loadDirectories({'vehicles/common/', path}) then -- important: the real vehicle dir is always last, all additional paths come first
      log('E', 'main', 'unable to load vehicle: aborted loading')
      return
    end
    table.insert(loadingTimes, {'1.X.X.X loadDirectories (sum)', hp1:stopAndReset()})

    -- experimental: vehicle file caching
    --[[
    local mp = require 'MessagePack'
    local f = io.open(vCacheFilename, 'w')
    if f then
      f:write(mp.pack(v.vehicles))
      f:close()
    end
    table.insert(loadingTimes, {'2. messagepack serialize', hp1:stopAndReset()})
    --]]
  --end

  -- you can change the data in here before it gets submitted to the physics

  -- submit to physics
  hp1:reset()
  v.pushToPhysics(obj)
  table.insert(loadingTimes, {'2.XX pushToPhysics (sum)', hp1:stopAndReset()})

  if v.data == nil then
    v.data = {}
  end

  initSystems()
  table.insert(loadingTimes, {'3 initSystems', hp1:stopAndReset()})

  -- temporary tire mark setting
  obj.slipTireMarkThreshold = 15

  -- update gravity and other things
  obj:queueGameEngineLua( 'vehicleSpawned('..tostring(obj:getID())..')' )

  if settings and settings.getValue('outgaugeEnabled') == true then
    extensions.loadModule("outgauge")
  end

  -- load the extensions at this point in time, so the whole jbeam is parsed already
  extensions.loadModulesInDirectory("lua/vehicle/extensions/auto")

  -- extensions that always load
  extensions.loadModule("skeleton")

  local totalTime = hp2:stopAndReset()
  table.insert(loadingTimes, {'4.X.X.X total (sum)', totalTime})

  --log('D', "default.init", "init done - vehicle loading took "..totalTime..' ms')

  --perf.disable()
  --perf.saveDataToCSV('vehicle_boottime_perf.csv')

  -- dump boot time stats
  --log('D', 'default.init', 'Vehicle boot time stats:')
  --for _, t in pairs(loadingTimes) do
  --  log('D', 'default.init', '  * ' .. rpad(t[1], 34, ' ') .. ' = ' .. lpad(string.format('%5.3f', t[2]), 8, ' ') .. ' ms')
  --end


  -- dump them into a file as well
  --[[
  local f = io.open('vehicle_boottime.csv', "w")
  if f then
    table.sort(loadingTimes, function(a, b) return a[2] > b[2] end)
    f:write(";name, time in ms\n")
    for _,vv in ipairs(loadingTimes) do
      f:write(vv[1] .. ', ' .. tostring(vv[2]) .. "\n")
    end
    f:close()
  end
  --]]

  extensions.hook('onVehicleLoaded', retainDebug)
end

-- various callbacks
function beamBroken(id, energy)
  beamstate.beamBroken(id, energy)
  wheels.beamBroke(id)
  powertrain.beamBroke(id)
  energyStorage.beamBroke(id)
  controller.beamBroke(id, energy)
end

-- only being called if the beam has deform triggers
function beamDeformed(id, ratio)
  beamstate.beamDeformed(id, ratio)
  controller.beamDeformed(id, ratio)
end

function couplerFound(nodeId, obj2id, obj2nodeId)
  -- print('couplerFound'..','..nodeId..','..obj2nodeId..','..obj2id)
  beamstate.couplerFound(nodeId, obj2id, obj2nodeId)
  controller.onCouplerFound(nodeId, obj2id, obj2nodeId)
  extensions.hook('onCouplerFound', nodeId, obj2id, obj2nodeId)
end

function couplerAttached(nodeId, obj2id, obj2nodeId)
  -- print('couplerAttached'..','..nodeId..','..obj2nodeId..','..obj2id)
  beamstate.couplerAttached(nodeId, obj2id, obj2nodeId)
  controller.onCouplerAttached(nodeId, obj2id, obj2nodeId)
  extensions.hook('onCouplerAttached', nodeId, obj2id, obj2nodeId)
end

function couplerDetached(nodeId, obj2id, obj2nodeId)
  -- print('couplerDetached'..','..nodeId..','..obj2nodeId..','..obj2id)
  beamstate.couplerDetached(nodeId, obj2id, obj2nodeId)
  controller.onCouplerDetached(nodeId, obj2id, obj2nodeId)
  extensions.hook('onCouplerDetached', nodeId, obj2id, obj2nodeId)
end

-- called when everything was cleared
function vehicleDestroy()
  --log('D', "default.vehicleDestroy", "vehicleDestroy()")
  -- when the vehicle gets unloaded, remove all sounds
  sounds.destroy()
  material.destroy()
end

-- called when the user pressed I
function vehicleResetted(retainDebug)
  local hp1 = HighPerfTimer()

  guihooks.reset()
  extensions.hook('onReset', retainDebug)
  ai.reset()
  partmgmt.reset()
  mapmgr.reset()
  scenario.reset()

  if not initCalled then
    --log('D', "default.vehicleResetted", "vehicleResetted()")
    wheels.reset()
    electrics.reset()
    powertrain.reset()
    energyStorage.reset()
    controller.reset()
    wheels.initSecondStage()
    controller.resetSecondStage()
    drivetrain.reset()
    props.reset()
    sensors.reset()
    if not retainDebug then
      bdebug.reset()
    end
    beamstate.reset()
    thrusters.reset()
    input.reset()
    hydros.reset()
    material.reset()
    fire.reset()
  end
  initCalled = false

  gui.message("", 0, "^damage\\.") -- clear damage messages on vehicle restart
  obj:queueGameEngineLua("extensions.hook('onVehicleResetted'," .. obj:getID() .. ")")
  loadingTimes['4_vehicleResetted'] = hp1:stopAndReset()
end

function nodeCollision(p)
  wheels.updateWheelSlip(p)
  fire.nodeCollision(p)
  particlefilter.nodeCollision(p)
end

function objectCollision(objId)
  beamstate.objectCollision(objId)
end

function setControllingPlayers(players)
  playerInfo.seatedPlayers = players
  playerInfo.anyPlayerSeated = tableSize(players) > 0
  playerInfo.firstPlayerSeated = players[0] ~= nil

  if playerInfo.anyPlayerSeated and controller and controller.mainController then controller.mainController.sendTorqueData() end
  bdebug.activated(playerInfo.anyPlayerSeated)
  ai.stateChanged()
  extensions.hook('activated', playerInfo.anyPlayerSeated) -- backward compatibility
  guihooks.trigger('VehicleFocusChanged', {id = obj:getID(), mode = playerInfo.anyPlayerSeated})
  -- TODO: clean below up ...
  obj:queueGameEngineLua("extensions.hook('onVehicleFocusChanged'," .. serialize({id = obj:getID(), mode = playerInfo.anyPlayerSeated}) .. ")")
end

function instabilityDetected()
  -- enable break debug, so we see what beams broke
  log('E', "default.instabilityDetected", "instability detected for vehicle " .. tostring(v.vehicleDirectory))
  ui_message({txt="vehicle.main.instability", context={vehicle=tostring(v.vehicleDirectory)}}, 10, 'instability')

  BeamEngine:setEnabled(false)
  obj:requestReset(RESET_PHYSICS)

   --if settings and settings.getValue('disableDynamicCollision') == true then
   --  Settings.dynamicCollision = false
   --end
end

function exportPersistentData()
  local s = serializePackages('reload')
  --log('D', "default.exportPersistentData", s)
  obj:setPersistentData(s)
end

function importPersistentData(s)
  --log('D', "default.importPersistentData", s)
  -- deserialize extensions first, so the extensions are loaded before they are trying to get deserialized
  deserializePackages(unserialize(s))
end

-- used by the GE camera code
function requestCameraConfig()
  if not cameraData then cameraData = {} end
  obj:queueGameEngineLua("extensions.hook('onVehicleCameraConfigChanged'," .. obj:getID() .. "," .. serialize(cameraData) .. ")")
end

function onSettingsChanged()
  extensions.hook('onSettingsChanged')
  controller.settingsChanged()
end

