local M = {}

local settings = require("simplesettings")

local max = math.max
local min = math.min
local random = math.random

local conversion = {
  kelvinToCelsius = -273.15,
  celsiusToKelvin = 273.15,
  avToRPM = 9.549296596425384,
}

local warningState = {}
local parentEngine = nil
local isSmoke = 0

--Thermal variables
M.engineBlockTemperature = 0
M.cylinderWallTemperature = 0
M.coolantTemperature = 0
M.oilTemperature = 0
M.exhaustTemperature = 0
M.radiatorFanSpin = 0
local engineBlockEnergyCoef = 0
local cylinderWallEnergyCoef = 0
local engineBlockMeltingTemperature = 0
local cylinderWallMeltingTemperature = 0
local oilEnergyCoef = 0
local coolantMass = 0
local coolantEnergyCoef = 0
local invInitialCoolantMass = 0
local exhaustEnergyCoef = 0
local burnEfficiencyCoef = 0
local engineBlockAirCoolingEfficiency = 0
local blockFanMaxAirSpeed = 0
local blockFanRPMCoef = 0

--Radiator
local radiatorFanType = nil
local radiatorFanMaxAirSpeed = 0
local radiatorFanTemperature = 0
local thermostatTemperature = 0
local oilThermostatTemperature = 0
local fanAirSpeed = 0
local mechanicalRadiatorFanRPMCoef = 0
local radiatorCoef = 0
local oilRadiatorCoef = 0
local hasCoolantRadiator = false
local hasOilRadiator = false
local electricalRadiatorFanSmoother = nil
local mechanicalFanRPMCoef = 0

--Damage
local headGasketDamageThreshold = 0
local pistonRingDamageThreshold = 0
local connectingRodDamageThreshold = 0
local engineBlockTemperatureDamageThreshold = 0
local cylinderWallTemperatureDamageThreshold = 0
M.engineBlockOverheatDamage = 0
M.oilOverheatDamage = 0
M.cylinderWallOverheatDamage = 0
M.headGasketBlown = false
M.pistonRingsDamaged = false
M.connectingRodBearingsDamaged = false
M.radiatorDamage = 0
M.engineBlockMelted = false
M.cylinderWallsMelted = false
local coolantOverpressureLeakRate = 0
local coolantHeadGasketLeakRate = 0
local radiatorLeakRate = 0
local radiatorDamageDeformGroup = "radiator_damage"
local radiatorDeformThreshold = 0
local oilStarvingTimer = 0
local oilStarvingTimerThreshold = 5
local oilStarvingDamage = 0

--Particles & Sound
local engineSteamParticleTick = 0
local exhaustSteamParticleTick = 0
local exhaustOilParticleTick = 0
local exhaustSmokeParticleTick = 0
local radiatorSteamParticleTick = 0
local knockSoundTick = 0
local particulates = 0
local idleParticulates = 0

local startPreHeated = true

--Thermal constants
local constants = {
  preHeatTemperature = 80,
  oilSpecHeat = 1800,
  coolantSpecHeat = 4000,
  exhaustSpecHeat = 500,
  minimumCoolantMass = 1.5,
  coolantTemperatureDamageThreshold = 120,
  maxCoolantTemperature = 130,
  oilTemperatureDamageThreshold = 150,
  exhaustCondensationThresholdEnvTemp = 15,
  exhaustCondensationThresholdBlockTemp = 60,
}

--Nodes
local coolantCapNodes = {}
local radiatorNodes = {}
local engineNodes = {}
local exhaustNodes = {}
local exhaustStartNode = nil
local exhaustBeams = nil
local exhaustTree = nil
local updateThermalsGFXMethod = nop
local updateMechanicsGFXMethod = nop

local function sendWarningStateToUi ()
  if not playerInfo.firstPlayerSeated then return end
  guihooks.trigger('WarningState', {module = 'engine', data = warningState})
end

local function updateThermalsGFX(dt)
  local tEnv = obj:getEnvTemperature() + conversion.kelvinToCelsius
  --k "spring" values
  local kExhaustToAir = 0.00001
  local kCylinderWallToCoolant = 48000
  local kCylinderWallToOil = 150
  local kOilToAir = 30
  local kCoolantToBlock = 20000
  local kCylinderWallToBlock = 10000

  local engineRPM = parentEngine.outputAV1 * conversion.avToRPM

  if hasCoolantRadiator then
    --get radiator damage (if there is any) and calculate coolant leak rate based on that
    M.radiatorDamage = max(beamstate.deformGroupDamage[radiatorDamageDeformGroup] and
      (beamstate.deformGroupDamage[radiatorDamageDeformGroup].damage - radiatorDeformThreshold) or 0, 0)

    if M.radiatorDamage > 0 and not warningState.radiator then
      gui.message("vehicle.engine.radiatorDamaged", 10, "damage.radiator")
      warningState.radiator = true
      sendWarningStateToUi()
    end
    radiatorLeakRate = M.radiatorDamage * 10
  end

  local radiatorFanRPM = 0
  if radiatorFanType == "electric" then
    --eletric fans are either on or off, depending on coolant temperature
    if M.coolantTemperature >= radiatorFanTemperature then
      fanAirSpeed = radiatorFanMaxAirSpeed
    elseif M.coolantTemperature <= thermostatTemperature * 1.05 then
      fanAirSpeed = 0
    end
    radiatorFanRPM = electricalRadiatorFanSmoother:getUncapped(fanAirSpeed > 0 and 2500 or 0, dt)
  elseif radiatorFanType == "mechanical" then
    --mechanical fans are tied to the RPM but at some point they won't go any faster (because they are linked with a clutch)
    radiatorFanRPM = engineRPM * mechanicalFanRPMCoef
    fanAirSpeed = radiatorFanMaxAirSpeed * min(engineRPM * mechanicalRadiatorFanRPMCoef, 1)
  end
  M.radiatorFanSpin = (M.radiatorFanSpin + radiatorFanRPM * dt) % 360

  local vel = vec3(obj:getVelocity())
  local direction = vec3(obj:getDirectionVector()):normalized()
  local airSpeedThroughVehicle = vel:dot(direction)

  --radiator is only actually used above a certain temperature
  local radiatorActive = min((hasCoolantRadiator and M.coolantTemperature > thermostatTemperature and not parentEngine.isDisabled) and
    M.coolantTemperature - thermostatTemperature or 0, 1)
  local oilRadiatorActive = min((hasOilRadiator and M.oilTemperature > oilThermostatTemperature and not parentEngine.isDisabled) and
    M.oilTemperature - oilThermostatTemperature or 0, 1)
  --Efficiency of the cooling system drops with decreasing coolant mass and raising temps above damage threshold
  local coolantEfficiency = (coolantMass > constants.minimumCoolantMass or M.coolantTemperature < constants.maxCoolantTemperature) and
  (coolantMass * invInitialCoolantMass * max(min((constants.maxCoolantTemperature - M.coolantTemperature) * 0.1, 1), 0)) or 0

  local underWaterBlockCoolingCoef = 1
  --if a node is underwater we want to increase the block to air cooling to simulate water cooling of the block
  for _,v in pairs(engineNodes) do
    underWaterBlockCoolingCoef = underWaterBlockCoolingCoef + (obj:inWater(v) and 1000 or 0)
  end

  --Step 1: Calculate the "forces" with our "spring" k values
  local currentEngineEfficiency = burnEfficiencyCoef[math.floor(parentEngine.engineLoad * 100) * 0.01]
  local burnEnergyPerUpdate = parentEngine.engineWorkPerUpdate * currentEngineEfficiency

  local energyToCylinderWall = 0.5 * burnEnergyPerUpdate + 0.5 * parentEngine.pumpingLossPerUpdate + parentEngine.frictionLossPerUpdate * 0.0
  local energyToExhaust = 0.5 * burnEnergyPerUpdate + 0.5 * parentEngine.pumpingLossPerUpdate
  local energyToOil = parentEngine.frictionLossPerUpdate

  local radiatorAirSpeed = max(airSpeedThroughVehicle * 0.7, fanAirSpeed) --reduce actual airspeed because the rad blocks part of the air
  local radiatorAirSpeedCoef = max(radiatorAirSpeed / (3 + radiatorAirSpeed), 0.01)

  local blockFanAirSpeedCoef = 1 + blockFanMaxAirSpeed * min(engineRPM * blockFanRPMCoef, 1)
  local cylinderWallToCoolantAirCooledCoef = blockFanMaxAirSpeed > 0 and 0 or 1 --kill the energy transfer from wall to coolant on an air cooled engine (no coolant)

  local energyCylinderWallToCoolant   = (M.cylinderWallTemperature - M.coolantTemperature) * kCylinderWallToCoolant * coolantEfficiency * cylinderWallToCoolantAirCooledCoef
  local energyCoolantToAir            = (M.coolantTemperature - tEnv) * radiatorCoef * radiatorActive * radiatorAirSpeedCoef * coolantEfficiency
  local energyCoolantToBlock          = (M.coolantTemperature - M.engineBlockTemperature) * kCoolantToBlock * coolantEfficiency * cylinderWallToCoolantAirCooledCoef
  local energyCylinderWallToBlock     = (M.cylinderWallTemperature - M.engineBlockTemperature) * kCylinderWallToBlock

  local energyCylinderWallToOil   = (M.cylinderWallTemperature - M.oilTemperature) * kCylinderWallToOil
  local energyOilSumpToAir        = (M.oilTemperature - tEnv) * kOilToAir
  local energyOilToAir            = (M.oilTemperature - tEnv) * oilRadiatorCoef * radiatorAirSpeedCoef * oilRadiatorActive
  local energyBlockToAir          = (M.engineBlockTemperature - tEnv) * blockFanAirSpeedCoef * engineBlockAirCoolingEfficiency * underWaterBlockCoolingCoef
  local exhaustTempDiff           = M.exhaustTemperature - tEnv
  local exhaustTempSquared        = exhaustTempDiff * exhaustTempDiff
  local energyExhaustToAir        = exhaustTempSquared * exhaustTempSquared * kExhaustToAir
  --local energyFireToBlock         = 0

  --We can lose coolant in different places, sum up all the rates and calculate the new mass
  local overallLeakRate = coolantOverpressureLeakRate + coolantHeadGasketLeakRate + radiatorLeakRate

  --Step 2: The integrator
  M.cylinderWallTemperature   = max(M.cylinderWallTemperature + (energyToCylinderWall - (energyCylinderWallToOil + energyCylinderWallToCoolant + energyCylinderWallToBlock) * dt) * cylinderWallEnergyCoef , tEnv)
  M.coolantTemperature        = min(max(M.coolantTemperature + (energyCylinderWallToCoolant - energyCoolantToAir - energyCoolantToBlock) * coolantEnergyCoef * dt, tEnv), constants.maxCoolantTemperature)
  M.oilTemperature            = max(M.oilTemperature + (energyToOil + (energyCylinderWallToOil - energyOilToAir - energyOilSumpToAir) * dt) * oilEnergyCoef, tEnv)
  M.engineBlockTemperature    = max(M.engineBlockTemperature + (energyCoolantToBlock + energyCylinderWallToBlock - energyBlockToAir) * engineBlockEnergyCoef * dt, tEnv)
  M.exhaustTemperature        = max(M.exhaustTemperature + (energyToExhaust - energyExhaustToAir * dt) * exhaustEnergyCoef , tEnv)
  coolantMass                 = max(coolantMass - overallLeakRate * dt, constants.minimumCoolantMass)

  local particleAirspeed      = electrics.values.airspeed
  local engineRunning         = (not (parentEngine.isDisabled or parentEngine.isStalled)) and 1 or 0
  engineSteamParticleTick     = engineSteamParticleTick > 1 and 0 or engineSteamParticleTick + dt * 4
  radiatorSteamParticleTick   = radiatorSteamParticleTick > 1 and 0 or radiatorSteamParticleTick + dt * (200 * M.radiatorDamage + 2 * particleAirspeed)
  exhaustSteamParticleTick    = exhaustSteamParticleTick > 1 and 0 or exhaustSteamParticleTick + dt * (0.01 * engineRPM + 0.2 * particleAirspeed) * engineRunning
  exhaustOilParticleTick      = exhaustOilParticleTick > 1 and 0 or exhaustOilParticleTick + dt * (0.01 * engineRPM + 0.1 * particleAirspeed) * engineRunning
  exhaustSmokeParticleTick    = exhaustSmokeParticleTick > 1 and 0 or exhaustSmokeParticleTick + dt * (0.02 + (0.02 * engineRPM + 0.02 * particleAirspeed)) * engineRunning

  --airspeed depending particle type selection
  local coolantHeavyParticleType  = particleAirspeed < 10 and 35 or 37
  local coolantLightParticleType  = particleAirspeed < 10 and 48 or 49
  local oilParticleType           = particleAirspeed < 10 and 36 or 38
  local steamParticleType         = particleAirspeed < 10 and 34 or 39
  local lightSmokeParticleType    = particleAirspeed < 10 and 40 or 41
  local heavySmokeParticleType    = particleAirspeed < 10 and 42 or 43
  local condensationParticleType  = particleAirspeed < 10 and 46 or 47
  local rand = random(1) --some random value for particle emission
  local particulateEmission = (particulates * parentEngine.engineLoad) + idleParticulates

  if M.coolantTemperature > constants.coolantTemperatureDamageThreshold then
    --our coolant is too hot, so our radiator cap is releasing pressure/steam and therefore coolant mass
    coolantOverpressureLeakRate = 0.01 --10g/s

    if engineSteamParticleTick > 1 then
      --emit steam as long as there is still coolant left
      obj:addParticleByNodesRelative(coolantCapNodes[1], coolantCapNodes[2], 1-rand , coolantHeavyParticleType, 0, 1)
    end

    if not warningState.coolant then
      gui.message("vehicle.engine.coolantOverheating", 10, "damage.coolant")
      warningState.coolant = true
      sendWarningStateToUi()
    end
  else
    --if the coolant cools down again, we don't leak anymore because of overpressure
    coolantOverpressureLeakRate = 0
    warningState.coolant = false
  end

  if M.engineBlockTemperature > engineBlockTemperatureDamageThreshold then
    M.engineBlockOverheatDamage = min(M.engineBlockOverheatDamage + parentEngine.engineWorkPerUpdate, headGasketDamageThreshold)
    if M.engineBlockOverheatDamage >= headGasketDamageThreshold and not M.headGasketBlown then
      --if we reach the headgasket damage threshold, we will lose more coolant
      M.headGasketBlown = true
      --without a working headgasket we don't have full compression anymore -> less torque
      parentEngine:scaleOutputTorque(0.8)
      coolantHeadGasketLeakRate = 0.1 --100g/s
      --let's get rid of a bit of coolant immediately
      coolantMass = coolantMass * 0.9
      gui.message("vehicle.engine.headGasketDamaged", 10, "damage.gasket")
      warningState.gasket = true
      sendWarningStateToUi()

      --implement nice steam "explosion" here
    end

    if M.engineBlockTemperature > engineBlockMeltingTemperature and not M.engineBlockMelted then
      parentEngine:scaleFriction(10000) --essentially kill the engine
      gui.message("vehicle.engine.blockMelted", 10, "damage.engineblock")
      M.engineBlockMelted = true
      warningState.engineblock = true
      sendWarningStateToUi()
    end
  end

  if M.oilTemperature > constants.oilTemperatureDamageThreshold then
    local diff = M.oilTemperature - constants.oilTemperatureDamageThreshold
    --increase engine friction relative to temperature of overheated oil
    local frictionCoef = 1 + diff * dt * 0.005
    parentEngine:scaleFriction(frictionCoef)
    M.oilOverheatDamage = min(M.oilOverheatDamage + parentEngine.engineWorkPerUpdate, connectingRodDamageThreshold)
    if M.oilOverheatDamage >= connectingRodDamageThreshold and not M.connectingRodBearingsDamaged then
      gui.message("vehicle.engine.rodBearingsDamaged", 10, "damage.rodbearings")

      M.connectingRodBearingsDamaged = true
      warningState.rodbearings = true
      sendWarningStateToUi()
    end

    if not warningState.oil then
      gui.message("vehicle.engine.oilOverheating", 10, "damage.oil")
      warningState.oil = true
      sendWarningStateToUi()
    end
  else
    warningState.oil = false
  end

  if M.cylinderWallTemperature > cylinderWallTemperatureDamageThreshold then
    M.cylinderWallOverheatDamage = min(M.cylinderWallOverheatDamage + parentEngine.engineWorkPerUpdate, pistonRingDamageThreshold)
    --if our cylinder wall gets too hot, the piston rings will be damaged eventually
    if M.cylinderWallOverheatDamage >= pistonRingDamageThreshold and not M.pistonRingsDamaged then
      gui.message("vehicle.engine.pistonRingsDamaged", 10, "damage.rings")
      M.pistonRingsDamaged = true
      --Damaged piston rings cause a loss of compression and therefore less torque
      parentEngine:scaleOutputTorque(0.8)
      warningState.rings = true
      sendWarningStateToUi()
    end

    if M.cylinderWallTemperature > cylinderWallMeltingTemperature and not M.cylinderWallsMelted then
      parentEngine:scaleFriction(10000) --essentially kill the engine
      gui.message("vehicle.engine.cylinderWallsMelted", 10, "damage.cylinderwalls")
      M.cylinderWallsMelted = true
      warningState.cylinderwalls = true
      sendWarningStateToUi()
    end
  end
  
  if M.radiatorDamage > 0 and radiatorSteamParticleTick > 1 and M.coolantTemperature >= 70 and hasCoolantRadiator then
    --emit steam from the radiator if it's damaged and we have coolant left
    if #radiatorNodes >= 2 then
      local coolantLeakParticle = M.coolantTemperature > 95 and coolantHeavyParticleType or coolantLightParticleType
      obj:addParticleByNodesRelative(radiatorNodes[1], radiatorNodes[2], rand, coolantLeakParticle, 0, 1)
	  obj:addParticleByNodesRelative(radiatorNodes[1], radiatorNodes[2], rand, coolantLeakParticle, 0, 1)	 --//////////////
	if combustionEngineThermals.isSmoke == 0 then                  											 --//////////////
		sounds.smokeOn()																					 --//////////////
		combustionEngineThermals.isSmoke = 1																 --//////////////
	end																										 --//////////////	
    end		
  end

  if M.headGasketBlown then
    --emit steam from the engine since we can't keep the coolant under control anymore
    if engineSteamParticleTick > 1 and hasCoolantRadiator then
      obj:addParticleByNodesRelative(engineNodes[2], engineNodes[1], rand * 3, coolantHeavyParticleType, 0, 1)
	if combustionEngineThermals.isSmoke == 0 then                  											 --//////////////
		sounds.smokeOn()																					 --//////////////
		combustionEngineThermals.isSmoke = 1																 --//////////////
	end																										 --//////////////
    end
  end

  if M.connectingRodBearingsDamaged then
    knockSoundTick = knockSoundTick > 1 and 0 or knockSoundTick + dt * engineRPM * 0.008333
    if knockSoundTick > 1 then
      --make a knocking sound if the bearings are damaged
      sounds.playSoundOnceAtNode("Knock", engineNodes[1], 0.1)
    end
  end

  --exhaust emission
  for _,n in pairs(exhaustNodes) do
    --regular exhaust smoke
    if exhaustSmokeParticleTick > 1 and particulateEmission > 0 then
      if particulateEmission > 0.05 and particulateEmission < 0.3 then
        obj:addParticleByNodesRelative(n.finish, n.start, engineRPM * -0.0004 - 3, lightSmokeParticleType, 0, 1)
      end
      if particulateEmission >= 0.3 then
        obj:addParticleByNodesRelative(n.finish, n.start, engineRPM * -0.0004 - 3, heavySmokeParticleType, 0, 1)
      end
    end

    --steam from broken head gasket
    if M.headGasketBlown and exhaustSteamParticleTick > 1 and coolantMass > constants.minimumCoolantMass and not (parentEngine.isDisabled or parentEngine.isStalled) and hasCoolantRadiator then
      --also emit steam from all exhaust ends because we are actually vaporizing coolant in the combustion chamber
      obj:addParticleByNodesRelative(n.finish, n.start, engineRPM * -0.0004 - 3, steamParticleType, 0, 1)
    end

    if tEnv <= constants.exhaustCondensationThresholdEnvTemp and M.engineBlockTemperature <= constants.exhaustCondensationThresholdBlockTemp and exhaustSteamParticleTick > 1 then
      obj:addParticleByNodesRelative(n.finish, n.start, engineRPM * -0.0004 - 1.3, condensationParticleType, 0, 1)
    end

    --blue smoke from broken piston rings
    if M.pistonRingsDamaged and exhaustOilParticleTick > 1 and not (parentEngine.isDisabled or parentEngine.isStalled) then
      --emit blue smoke from all exhaust ends because we are burning oil with damaged piston rings
      obj:addParticleByNodesRelative(n.finish, n.start, engineRPM * -0.0004 - 2, oilParticleType, 0, 1)
    end
  end

  if gui.updateStreams and streams.engineThermalData > 0 then
    gui.send('engineThermalData', {
        coolantTemperature = M.coolantTemperature,
        oilTemperature = M.oilTemperature,
        engineBlockTemperature = M.engineBlockTemperature,
        cylinderWallTemperature = M.cylinderWallTemperature,
        exhaustTemperature = M.exhaustTemperature,
        radiatorAirSpeed = radiatorAirSpeed,
        radiatorAirSpeedEfficiency = radiatorAirSpeedCoef,
        fanActive = fanAirSpeed > 0,
        thermostatStatus = radiatorActive,
        oilThermostatStatus = oilRadiatorActive,
        coolantLeakRate = overallLeakRate,
        coolantEfficiency = coolantEfficiency,
        engineEfficiency = 1 / (currentEngineEfficiency + 1),
        energyToCylinderWall = energyToCylinderWall,
        energyToOil = energyToOil,
        energyToExhaust = energyToExhaust,
        energyCoolantToAir = energyCoolantToAir * dt,
        energyCylinderWallToCoolant = energyCylinderWallToCoolant * dt,
        energyCoolantToBlock = energyCoolantToBlock * dt,
        energyCylinderWallToBlock = energyCylinderWallToBlock * dt,
        energyCylinderWallToOil = energyCylinderWallToOil * dt,
        energyOilToAir = energyOilToAir * dt,
        energyOilSumpToAir = energyOilSumpToAir * dt,
        energyBlockToAir = energyBlockToAir * dt,
        energyExhaustToAir = energyExhaustToAir * dt,
        engineBlockOverheatDamage = M.engineBlockOverheatDamage,
        oilOverheatDamage = M.oilOverheatDamage,
        cylinderWallOverheatDamage = M.cylinderWallOverheatDamage,
        headGasketBlown = M.headGasketBlown,
        pistonRingsDamaged = M.pistonRingsDamaged,
        connectingRodBearingsDamaged = M.connectingRodBearingsDamaged,
        engineBlockMelted = M.engineBlockMelted,
        cylinderWallsMelted = M.cylinderWallsMelted,
        thermostatTemperature = thermostatTemperature,
        oilThermostatTemperature = oilThermostatTemperature,
      })
  end
end

local function updateMechanicsGFX(dt)
  --check if our oil is still in the oil sump or if it's forced towards the top of the sump
  local oilStarvingSeverness = min(max(sensors.gz2, -9.81), 9.81) * 0.1019368
  oilStarvingTimer = min(max(oilStarvingTimer + oilStarvingSeverness * dt, 0), oilStarvingTimerThreshold)

  if oilStarvingTimer >= oilStarvingTimerThreshold and not (parentEngine.isDisabled or parentEngine.isStalled) then
    local engineRPM = parentEngine.outputAV1 * conversion.avToRPM
    oilStarvingDamage = oilStarvingDamage + 0.00000001 * dt * engineRPM
    parentEngine:scaleFriction(1 + oilStarvingDamage)

    local oilParticleType = electrics.values['airspeed'] < 10 and 36 or 38

    for _,n in pairs(exhaustNodes) do
      if exhaustOilParticleTick > 1 and not parentEngine.isDisabled then
        --emit blue smoke from all exhaust ends because we are burning oil with damaged piston rings
        obj:addParticleByNodesRelative(n.finish, n.start, engineRPM * -0.0004 - 2, oilParticleType, 0, 1)
      end
    end

    if not warningState.oilstarvation then
      gui.message("vehicle.engine.starvedOfOil", 10, "damage.oilstarvation")
      warningState.oilstarvation = true
      sendWarningStateToUi()
    end
  else
    warningState.oilstarvation = false
  end
end

local function getExhaustEndNodes(startNode, exhaustTree)
  local branch = exhaustTree.children[startNode]
  local endNodes = {}

  for k,child in pairs(branch.children) do
    --if at this point something broke away or we reached the original end of the branch
    if child.childrenCount ~= child.initialChildrenCount or child.initialChildrenCount == 0 then
      --save the nodes as exit nodes
      table.insert(endNodes, {start = child.previous, finish = child.cid})
    end
    --if we have children left and we are not broken
    if child.childrenCount > 0 and not child.isBroken then
      --continue to search for more exit nodes in this branch
      arrayConcat(endNodes, getExhaustEndNodes(k, branch))
    end
  end

  return endNodes
end

local function parseExhaustTree(currentBranch, exhaustBeams)
  --copy table to not mess with the original one
  local beams = shallowcopy(exhaustBeams)
  local currentBeamKey, currentBeam = next(beams, nil)
  currentBranch.childrenCount = 0
  currentBranch.children = {}

  while currentBeam ~= nil do
    --if your node connects to the current beam (via node1 or node2)
    if currentBranch.cid == currentBeam.id1 then
      beams[currentBeamKey] = nil --beam is handled, ignore further down the line
      --build child branch based on current beam
      currentBranch.children[currentBeam.id2] = parseExhaustTree({cid = currentBeam.id2, previous = currentBranch.cid, beam = currentBeam.cid}, beams)
      currentBranch.childrenCount = currentBranch.childrenCount + 1
    elseif currentBranch.cid == currentBeam.id2 then
      --same as above but with switched id1 <-> id2
      beams[currentBeamKey] = nil
      currentBranch.children[currentBeam.id1] = parseExhaustTree({cid = currentBeam.id1, previous = currentBranch.cid, beam = currentBeam.cid}, beams)
      currentBranch.childrenCount = currentBranch.childrenCount + 1
    end

    currentBeamKey, currentBeam = next(beams, currentBeamKey)
  end

  currentBranch.initialChildrenCount = currentBranch.childrenCount
  return currentBranch
end

local function buildExhaustTree()
  exhaustStartNode = nil
  exhaustBeams = {}
  local exhaustBeamCache = {}

  --search for the exhaust start node
  for _,n in pairs(v.data.nodes) do
    if n.isExhaust then
      if exhaustStartNode ~= nil and exhaustStartNode ~= n then
        log("E", "engine.buildExhaustTree", "Found multiple exhaust start nodes, please use only one. Old node: "..exhaustStartNode.cid..", new node: "..n.cid)
        return false
      end
      exhaustStartNode = n
    end
  end

  if not exhaustStartNode then
    log("E", "engine.buildExhaustTree", "No exhaust start node specified")
    return false
  end

  --find all exhaust beams
  for _,b in pairs(v.data.beams) do
    if b.isExhaust then
      --one table for immediate use
      table.insert(exhaustBeamCache, b)
      --one table for look ups when a beam breaks
      exhaustBeams[b.cid] = true
    end
  end

  exhaustTree = {children = {}}
  --build exhaust tree recursively
  exhaustTree.children[exhaustStartNode.cid] = parseExhaustTree({cid = exhaustStartNode.cid}, exhaustBeamCache)

  --find initial exhaust end points
  exhaustNodes = getExhaustEndNodes(exhaustStartNode.cid, exhaustTree)
  if #exhaustNodes <= 0 then
    log("E", "engine.buildExhaustTree", "No exhaust end nodes found")
    return false
  end

  return true
end

local function exhaustBeamBroken(id, exhaustTree)
  for _,v in pairs(exhaustTree.children) do
    --if the broken beam matches one of our tree beams
    if v and v.beam == id then
      --break off this branch
      v.isBroken = true
      exhaustTree.childrenCount = exhaustTree.childrenCount - 1
    elseif v and v.children then
      exhaustBeamBroken(id, v)
    end
  end
end

local function beamBroke(id)
  if exhaustBeams and exhaustBeams[id] then
    exhaustBeams[id] = false
    --break off a tree branch
    exhaustBeamBroken(id, exhaustTree)
    --and find the new exit nodes
    exhaustNodes = getExhaustEndNodes(exhaustStartNode.cid, exhaustTree)
  end
end

local function initThermals(jbeamData)
  M.updateThermalsGFX = nop
  M.updateMechanicsGFX = nop

  local tEnv = obj:getEnvTemperature() + conversion.kelvinToCelsius
  startPreHeated = settings.getValue("startThermalsPreHeated")
  --default temperatures, can be adjusted to fit whatever our goal is (starting up with cold vs warm car)
  local startingTemperature = startPreHeated and constants.preHeatTemperature or tEnv
  M.engineBlockTemperature = startingTemperature
  M.cylinderWallTemperature = startingTemperature
  M.oilTemperature = startingTemperature
  M.coolantTemperature = startingTemperature
  M.exhaustTemperature = startingTemperature
  if not jbeamData.thermalsEnabled then
    --disable the whole thing unless stated otherwise
    --log("D", "engine.initThermals", "Engine thermals are disabled since they are missing in JBeam")
    return
  end

  fanAirSpeed = 0
  M.engineBlockOverheatDamage = 0
  M.oilOverheatDamage = 0
  M.cylinderWallOverheatDamage = 0
  M.headGasketBlown = jbeamData.headGasketBlownOverride or false
  M.pistonRingsDamaged = jbeamData.pistonRingsDamagedOverride or false
  M.connectingRodBearingsDamaged = jbeamData.connectingRodBearingsDamagedOverride or false
  M.radiatorDamage = 0
  M.cylinderWallsMelted = false
  M.engineBlockMelted = false
  M.radiatorFanSpin = 0
  coolantOverpressureLeakRate = 0
  coolantHeadGasketLeakRate = 0
  radiatorLeakRate = 0
  oilStarvingTimer = 0
  oilStarvingDamage = 0
  engineSteamParticleTick = 0
  exhaustSteamParticleTick = 0
  exhaustOilParticleTick = 0
  exhaustSmokeParticleTick = 0
  knockSoundTick = 0

  particulates = jbeamData.particulates or 0
  idleParticulates = jbeamData.idleParticulates or 0

  --default to some little coolant mass to prevent any divide by 0 issues
  coolantMass = jbeamData.coolantVolume or constants.minimumCoolantMass
  invInitialCoolantMass = 1 / coolantMass
  local oilMass = (jbeamData.oilVolume or 5) * 0.86
  burnEfficiencyCoef = {}
  for k,v in pairs(parentEngine.burnEfficiencyTable) do
    burnEfficiencyCoef[k] =  (1 / v) - 1
  end
  mechanicalRadiatorFanRPMCoef = 1 / (parentEngine.maxRPM * 0.7)

  local engineBlockMaterial = jbeamData.engineBlockMaterial or "iron"
  local engineBlockSpecHeat = 0
  local cylinderWallSpecHeat = 0
  if engineBlockMaterial == "iron" then
    engineBlockSpecHeat = 450
    cylinderWallSpecHeat = 450
    engineBlockMeltingTemperature = 1100
    cylinderWallMeltingTemperature = 1200
  elseif engineBlockMaterial == "aluminium" or engineBlockMaterial == "aluminum" then
    engineBlockSpecHeat = 910
    cylinderWallSpecHeat = 910
    engineBlockMeltingTemperature = 660
    cylinderWallMeltingTemperature = 700
  else
    log("E", "engine.initThermals", "Unknown engine block material specified: "..engineBlockMaterial)
    log("E", "engine.initThermals", "Engine thermals are disabled due to above error")
    return
  end

  engineBlockAirCoolingEfficiency = jbeamData.engineBlockAirCoolingEfficiency or 3
  blockFanMaxAirSpeed = jbeamData.blockFanMaxAirSpeed or 0
  blockFanRPMCoef = 1 / parentEngine.maxRPM
  radiatorFanType = jbeamData.radiatorFanType
  local radiatorArea = jbeamData.radiatorArea or 0
  radiatorFanMaxAirSpeed = jbeamData.radiatorFanMaxAirSpeed or 0
  mechanicalFanRPMCoef = jbeamData.mechanicalFanRPMCoef or 0.5
  local radiatorEffectiveness = jbeamData.radiatorEffectiveness or 0
  radiatorFanTemperature = jbeamData.radiatorFanTemperature or 110
  thermostatTemperature = jbeamData.thermostatTemperature or 90
  oilThermostatTemperature = jbeamData.oilThermostatTemperature or 110
  local oilRadiatorArea = jbeamData.oilRadiatorArea or 0
  local oilRadiatorEffectiveness = jbeamData.oilRadiatorEffectiveness or 0
  cylinderWallTemperatureDamageThreshold = jbeamData.cylinderWallTemperatureDamageThreshold or 160
  headGasketDamageThreshold = jbeamData.headGasketDamageThreshold or 2000000
  pistonRingDamageThreshold = jbeamData.pistonRingDamageThreshold or 2000000
  connectingRodDamageThreshold = jbeamData.connectingRodDamageThreshold or 2000000
  engineBlockTemperatureDamageThreshold =  jbeamData.engineBlockTemperatureDamageThreshold or 140
  radiatorDeformThreshold = jbeamData.radiatorDeformThreshold or 0.015
  hasCoolantRadiator = radiatorArea > 0 and radiatorEffectiveness > 0
  hasOilRadiator = oilRadiatorArea > 0 and oilRadiatorEffectiveness > 0

  if radiatorFanType and radiatorFanType ~= "electric" and radiatorFanType ~= "mechanical" then
    log("E", "engine.initThermals", "Unknown radiator fan type specified: "..radiatorFanType)
  end
  electricalRadiatorFanSmoother = newTemporalSmoothing(500,1000)

  local engineBlockMass = 0

  if not jbeamData.engineBlock or not jbeamData.engineBlock._engineGroup_nodes then -- little hack to make it not fail while the jbeam parsing is broken
    log("E", "engine.initThermals", "No engineBlock node group specified")
    log("E", "engine.initThermals", "Engine thermals are disabled due to above error")
    return
  end
  for _,n in pairs(jbeamData.engineBlock._engineGroup_nodes) do
    engineBlockMass = engineBlockMass + v.data.nodes[n].nodeWeight
  end

  local cylinderWallMass = math.max(engineBlockMass / 100, 4)
  local exhaustMass = math.max(engineBlockMass / 10, 5)

  coolantEnergyCoef = 1 / (coolantMass * constants.coolantSpecHeat)
  oilEnergyCoef = 1 /  (oilMass * constants.oilSpecHeat)
  cylinderWallEnergyCoef = 1 / (cylinderWallMass * cylinderWallSpecHeat)
  engineBlockEnergyCoef = 1 / (engineBlockMass * engineBlockSpecHeat)
  exhaustEnergyCoef = 1 / (constants.exhaustSpecHeat * exhaustMass)
  radiatorCoef = radiatorEffectiveness * radiatorArea
  oilRadiatorCoef = oilRadiatorEffectiveness * oilRadiatorArea

  coolantCapNodes = {}
  radiatorNodes = {}
  engineNodes = {}

  if hasCoolantRadiator and not jbeamData.radiator then
    log("E", "engine.initThermals", "No radiator node group specified")
    log("E", "engine.initThermals", "Engine thermals are disabled due to above error")
    return
  else
    arrayConcat(radiatorNodes, jbeamData.radiator._engineGroup_nodes or {})
  end

  arrayConcat(coolantCapNodes, jbeamData.engineBlock._engineGroup_nodes or {})
  arrayConcat(engineNodes, jbeamData.engineBlock._engineGroup_nodes or {})

  if #coolantCapNodes < 2 then
    log("D", "engine.initThermals", "Wrong number of coolant cap nodes found. Should be at least 2, is: "..#coolantCapNodes)
  end

  if #engineNodes < 2 then
    log("E", "engine.initThermals", "Wrong number of engine nodes found. Should be at least 2, is: "..#engineNodes)
    log("E", "engine.initThermals", "Engine thermals are disabled due to above error")
    return
  end

  if hasCoolantRadiator and #radiatorNodes < 2 then
    log("D", "engine.initThermals", "Wrong number of radiator nodes found. Should be at least 2, is: "..#radiatorNodes)
  end

  if not buildExhaustTree() then
    log("E", "engine.initThermals", "Building the exhaust tree failed, please look above for the actual reason")
    log("E", "engine.initThermals", "Engine thermals are disabled due to above error")
    return
  end

  updateThermalsGFXMethod = updateThermalsGFX
  updateMechanicsGFXMethod = updateMechanicsGFX
end

local function updateGFX(dt)
  updateThermalsGFXMethod(dt)
  updateMechanicsGFXMethod(dt)
end

local function init(engine, jbeamData)
  parentEngine = engine
  initThermals(jbeamData)
  warningState = {}
  sendWarningStateToUi()
end

M.init = init
M.reset = init
M.isSmoke = isSmoke
M.updateGFX = updateGFX
M.beamBroke = beamBroke
M.getWarningStateForUi = sendWarningStateToUi

return M