-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.engineNode = nil

local windSound = nil
local wheelsSounds = nil

local soundBank = { sounds = {} }
local sfxprofilecounter = 0 -- local counter to enumerate the profiles without collisions, do not reset ever

local beamSounds = {}

M.uiDebugging = false

local soundObj = {}
soundObj.__index = soundObj

local totalSounds = 0
local playingSounds = 0

function newSoundObj(sndObj)
  if sndObj == nil then return nil end
  local data = {obj = sndObj, lastVol = 0, lastPitch = 0}
  setmetatable(data, soundObj)
  return data
end

function soundObj:setVolumePitch(vol, pitch)
  totalSounds = totalSounds + 1
  if vol < 0.05 then vol = 0 end
  if vol == 0 then
    if self.lastVol == 0 then return end
  else
    playingSounds = playingSounds + 1
    if math.abs(vol - self.lastVol) < 0.05 and math.abs(pitch - self.lastPitch) < 0.001 then return end
  end
  self.lastVol = vol
  self.lastPitch = pitch
  obj:setVolumePitch(self.obj, vol, pitch)
end

local function playSoundOnceAtNode(soundName, nodeID, volume)
  if volume < 0.1 then return end
  obj:playSFXOnce(soundName, nodeID, volume, 0)
end

function lerp(from,to,t)
  return from + (to - from) * math.min(math.max(0, t),1)
end

function invertLerp(from,to,value)
  value = math.min(math.max(from, value),to)
  return (value - from) /(to-from)
end

local function getSourceValue(sourcename)
  --in the future possibly replace with the same system props uses for source
  if sourcename == "gear" then return drivetrain.gear end
  if electrics.values[sourcename] ~= nil then return electrics.values[sourcename] end
  return nil
end

local function getSoundModifier(modName)
  local modifier = soundBank.modifiersNamed[modName]
  if modifier == nil then return 1 end

  local mVal = getSourceValue(modifier.source)
  if mVal == nil then return 1 end

  return math.min(math.max(modifier.min, modifier.factor * (mVal + modifier.offset)), modifier.max)
end

local function getNodeByName(nodename)
  if not v.data or not v.data.nodes then return nil end
  for _,node in pairs(v.data.nodes) do
    if node.name == nodename then return node.cid end
  end
  return nil
end

local function createSFXSource(filename, description, SFXProfileName, nodeID)
  local snd = obj:createSFXSource(filename, description, SFXProfileName, nodeID)
  if snd == nil then
    M.update = nop
    M.playSoundOnceAtNode = nop
    log('W', 'sounds.createSFXSource', 'failed to create sfx source: ' .. SFXProfileName.. ' from file ' .. filename .. ' with description ' .. description )
    return nil
  end
  return snd
end

local function createSoundObj(filename, description, SFXProfileName, nodeID)
  local snd = obj:createSFXSource(filename, description, SFXProfileName, nodeID)
  if snd == nil then
    M.update = nop
    M.playSoundOnceAtNode = nop
    log('W', 'sounds.createSoundObj', 'failed to create sfx source: ' .. SFXProfileName.. ' from file ' .. filename .. ' with description ' .. description )
    return nil
  end
  return newSoundObj(snd)
end
local ok = 1  --////////////
local ok1 = 1 --////////////
local function update(dt)
  totalSounds = 0
  playingSounds = 0
  -- sound bank
  for sndkey, snd in pairs(soundBank.sounds) do
    --skip if inavtive
    if snd.active == false then goto continue end

    local val = getSourceValue(snd.source) or 0
    val = snd.factor * (val + snd.offset )
    snd.lastVal = val

    local sndVol = 0
    local sndPitch = 0

    --check volume conditions
    if val < snd.volumeBlendInStartValue or val > snd.volumeBlendOutEndValue then
      sndVol = snd.minVolume
    end
    if val > snd.volumeBlendInStartValue and val < snd.volumeBlendInEndValue then
      --blend in volume
      sndVol = lerp(snd.minVolume, snd.maxVolume, (val - snd.volumeBlendInStartValue) / (snd.volumeBlendInEndValue - snd.volumeBlendInStartValue))
    elseif val > snd.volumeBlendInEndValue and val < snd.volumeBlendOutStartValue then
      sndVol = snd.maxVolume
    elseif val > snd.volumeBlendOutStartValue and val < snd.volumeBlendOutEndValue then
      --blend out volume
      sndVol = lerp(snd.minVolume, snd.maxVolume, (((val - snd.volumeBlendOutStartValue) / (snd.volumeBlendOutEndValue - snd.volumeBlendOutStartValue)) -1) * -1)
    end

    --check pitch conditions
    if val < snd.pitchBlendInStartValue then
      sndPitch = snd.minPitch
    end
    if val > snd.pitchBlendInEndValue then
      sndPitch = snd.maxPitch
    end
    if val > snd.pitchBlendInStartValue and val < snd.pitchBlendInEndValue then
      --blend pitch
      sndPitch = lerp(snd.minPitch, snd.maxPitch, (val - snd.pitchBlendInStartValue) / (snd.pitchBlendInEndValue - snd.pitchBlendInStartValue))
    end

    --apply modifiers if applicable
    for _,s in pairs(snd.volumeModifiers) do
      sndVol = sndVol * getSoundModifier(s)
    end
    for _,s in pairs(snd.pitchModifiers) do
      sndPitch = sndPitch * getSoundModifier(s)
    end

    snd.lastVolume = sndVol
    snd.lastPitch = sndPitch
    snd.clip:setVolumePitch(sndVol, sndPitch)
    ::continue::
  end

  -- beam sounds
  ----[[
  for bmkey, snd in pairs(beamSounds) do
    local currentStress = snd.smoothing:get(obj:getBeamStress(snd.beam), dt)                        -- find the stress on the current sound beam
    local impulse = (snd.lastStress - currentStress) / (dt * 30)                                -- take the difference in beam stress between this frame and the last frame, and save it as impulse
    snd.resonance = math.max(math.min(snd.resonance or 0, snd.maxStress), 0)  --limit sound factor to maxStress to prevent overly loud/long sounds
    local linDecay = (1 - snd.decayMode) * snd.resonance * snd.decayFactor * dt
    local expDecay = snd.decayMode * snd.resonance * snd.decayFactor * dt
    local factor = math.max(impulse, math.max(snd.resonance - linDecay - expDecay, 0)) --sound decays to create a smooth fade out. Rate is dependent on simulation speed.

    snd.lastStress = currentStress  --reset for next frame comparison
    --lastImpulse = impulse   --reset for next frame comparison -- unused?

    local volume = (factor * snd.volumeFactor * 1.3) / snd.maxStress  -- normalize volume (cancel out maxStress factor)
    local pitch = 1 + (snd.pitchFactor * (factor / snd.maxStress))  -- loud suspension sounds also gain a higher pitch

    if snd.resetTimer > 1 then  --prevent loud sounds from playing on spawn
      snd.clip:setVolumePitch(volume, pitch)
    else
      snd.resetTimer = snd.resetTimer + dt
      snd.clip:setVolumePitch(0, 0)
    end

    snd.resonance = factor --reset for next frame comparison
  end
  --]]

  -- wind
  if windSound then
    local speed = obj:getAirSpeed() -- speed against wind
    local vol = (speed * speed * 0.001)
    local pitch = speed / 60
    if vol > 1 then vol = 1 end
    windSound:setVolumePitch(vol, pitch)
  end

  -- wheels
  for wi,wd in pairs(wheels.wheels) do
    local w = wd.obj
    local slip = wd.lastSlip
    local slipEnergy = wd.slipEnergy
    local slipEnergySmLast = 0

    local skidVol = 0
    local skidPitch = 0
    local rollVol = 0
    local rollPitch = 0
	local rollPitch1 = 0 --////////////
	local rollVol1 = 0   --////////////

    if wd.contactMaterialID1 == 10 and wd.contactMaterialID2 == 4 and wd.contactDepth == 0  then
      -- cross fade smoothly between skid and rolling states (instead of the noticeable on/off toggling that was used in the past)
      local slipEnergySm = wd.slipEnergySmoother:getWithRateUncapped(slipEnergy, dt, 20)
      local energyLimit = 2 -- further than this value won't produce any more fade into rolling sound, so prevent values from going beyond this (prevent lag in fading between the two states)
      if slipEnergySm > energyLimit then wd.slipEnergySmoother :set(energyLimit) end
      local skidCrossfade =   math.max(0,math.min(slipEnergySm         ,1))
      local rollCrossfade = 1-math.max(0,math.min(slipEnergySm*0.3-0.25,1)) -- roll depends on high speed, let it sound a bit further than skid

      -- rolling sound
      local wheelSpeed = math.abs(w.angularVelocity * wd.radius)
      rollVol = rollCrossfade * math.min(math.sqrt(wheelSpeed * 0.018), 1)
      rollPitch = wheelSpeed * 0.125

      -- scrubbing/skidding sound, volume
      local slipEnergySmSkidVol = wd.slipEnergySkidVolSmoother:getWithRateUncapped(slipEnergy, dt, 200)
      local skidVolLimit = 37 -- further than this value won't make sounds any louder, so prevent values from going far beyond this (prevent volume lag)
      if slipEnergySmSkidVol > skidVolLimit then wd.slipEnergySkidVolSmoother:set(skidVolLimit) end
      skidVol = skidCrossfade * math.min(slipEnergySmSkidVol * 0.025, 1)

      -- scrubbing/skidding sound, pitch
      local slipEnergySmSkidPitch = wd.slipEnergySkidPitchSmoother:getWithRateUncapped(slipEnergy, dt, 1000)
      local skidPitchLimit = 180 -- further than this value won't change the pitch anymore, so prevent values from going far beyond this (prevent pitch lag)
      if slipEnergySmSkidPitch > skidPitchLimit then wd.slipEnergySkidPitchSmoother:set(skidPitchLimit) end
      local x = slipEnergySmSkidPitch
      -- blend from high freq skid squealing, to lower freq tire scrubbing
      local skidPitchCrossfade = math.max(0,math.min(1,-5+0.1*x))
      local scrubPitchCrossfade = 1-skidPitchCrossfade
      skidPitch = scrubPitchCrossfade*(x*0.001+0.7) + skidPitchCrossfade*(math.min(1.8,x*0.006+0.85))

      --if wi == 0 then log("I", "", "energ: "..graphs(slipEnergySmSkidVol/2, 35).." vol: "..graphs(skidVol*10, 20).." vol: "..dumps(skidVol).." energ: "..dumps(slipEnergySmSkidVol)) end
    end

    wheelsSounds[wi]["SkidTestSound"]:setVolumePitch(skidVol, skidPitch)
    wheelsSounds[wi]["RollingTestSound"]:setVolumePitch(rollVol, rollPitch)
	
	
--Siren & horn Mod	
		if skidVol > 0.3 then
		ok1 = ok1 + 1
		if ok1 > 300 and ok1 < 1000 then

		  obj:addParticleByNodesRelative(wd.node1, wd.node2, 1-math.random(1) , 35, 1, 1) end
		  
		if ok1 > 1000 and ok1 < 1700 then
		  obj:addParticleByNodesRelative(wd.node1, wd.node2, 1-math.random(1) , 37, 1, 2)
		  obj:addParticleByNodesRelative(wd.node1, wd.node2, 1-math.random(1) , 35, 1, 1) end
		  
		if ok1 > 1700 then
		  obj:addParticleByNodesRelative(wd.node1, wd.node2, 1-math.random(1) , 37, 1, 2)
		  obj:addParticleByNodesRelative(wd.node1, wd.node2, 1-math.random(1) , 37, 1, 4)
		  obj:addParticleByNodesRelative(wd.node1, wd.node2, 1-math.random(1) , 35, 1, 2) end
		  
		  
		elseif skidVol < 0.3 and ok1 ~= 1 then ok1 = ok1 - 1
		end
	

  if wd.contactMaterialID1 == 18 or wd.contactMaterialID1 == 15 then
	--if wd.contactMaterialID2 == 4 and wd.contactDepth == 0 then
		
      local slipEnergySm = wd.slipEnergySmoother:getWithRateUncapped(slipEnergy, dt, 20)
      local energyLimit = 2
      if slipEnergySm > energyLimit then wd.slipEnergySmoother :set(energyLimit) end
      local rollCrossfade = 1-math.max(0,math.min(slipEnergySm*0.3-0.25,1))
      local wheelSpeed = math.abs(w.angularVelocity * wd.radius)
      rollVol1 = rollCrossfade * math.min(math.sqrt(wheelSpeed * 0.1), 2)
      rollPitch1 = wheelSpeed * 0.1
	  rollgravel:setVolumePitch(rollVol1, rollPitch1)	
	  ok = 1
	--end	
  end
  
  	if ok == 50 then 
	 rollgravel:setVolumePitch(0, 1)
	else ok = ok + 1 end
--Siren & horn Mod	
	
  end

  if M.uiDebugging and playerInfo.firstPlayerSeated then
    guihooks.trigger("AudioDebug", soundBank)
  end

  -- print('s: '..playingSounds..'/'..totalSounds)
  
end

-- public interface
M.update  = update
M.playSoundOnceAtNode = playSoundOnceAtNode

local function addWheelSound(wheelID, wd, filename, description, profile)
  if wheelsSounds == nil then wheelsSounds = {} end
  if wheelsSounds[wheelID] == nil then wheelsSounds[wheelID] = {} end

  -- const char *filename, const char *descriptionName, const char* sfxProfileName, bool preload
  wheelsSounds[wheelID][profile] = createSoundObj(filename, description, profile, wd.node1)
end

local function loadSoundFiles(directory)
  --log('D', "sounds.loadSoundFiles", "loading sound files from: "..directory)
  local files = FS:findFiles(directory, "*.sbeam", -1, true, false)
  if not files or #files == 0 then
    --log('D', 'sounds.loadSoundFiles', 'unable to open directory for reading: ' .. directory)
    return
  end

  -- first: figure out all the filenames. TODO: recursive?
  local file = nil
  local sbeamFiles = {}
  for _, file in ipairs(files) do
    table.insert(sbeamFiles, file)
  end

  --load and merge
  local soundBank = {}
  for _, sbfn in pairs(sbeamFiles) do
    local tmp = readDictJSONTable(sbfn)
    if tmp == nil then
      log('E', 'sounds.lua', 'sbeam file empty or unable to parse: ' .. sbfn)
      goto continue
    end

    tableMergeRecursive(soundBank, tmp)
    ::continue::
  end

  -- fallback if no sounds were loaded
  if not soundBank.sounds then soundBank.sounds = {} end

  -- create lookup table
  if soundBank.modifiers then
    soundBank.modifiersNamed = {}
    for _, sbm in pairs(soundBank.modifiers) do
      soundBank.modifiersNamed[sbm.name] = sbm
    end
  end

  if type(soundBank.sounds) == 'table' then
    --log('D', "sounds.loadSoundFiles", 'loaded '.. #soundBank.sounds .. ' sounds from directory ' .. directory)
  else
    log('D', "sounds.loadSoundFiles", 'no sounds loaded from directory ' .. directory)
    return nil
  end

  return soundBank
end

local function checkLocalFile(folder,file)
  if not FS:fileExists(file) then
    local testfn = folder..file
    if FS:fileExists(testfn) then
      return testfn
    end
  end
  return file
end

local function getNextProfile()
  sfxprofilecounter = sfxprofilecounter + 1
  return 'LuaSoundProfile'..sfxprofilecounter..'_'.. os.time()
end

local function init()
  obj:deleteSFXSources()
  local cameraNode = 0
  if v.data.camerasInternal ~= nil then
    local k, v = next(v.data.camerasInternal)
    if v ~= nil then
      cameraNode = v.camNodeID
    end
  end

  --replace constants
  local maxrpm = 1
  M.engineNode = cameraNode
  if #powertrain.engineData > 0 then
    maxrpm = powertrain.engineData[1].maxRPM or 1

    -- Try to get a node on the engine. There is currently a libbeamng bug that
    -- causes sound sources at the camera to cause problems for multichannel
    -- setups, as small variations in position can cause sound to dither back
    -- and forth between the channels when in cockpit view.
    if powertrain.engineData[1].torqueReactionNodes then
      local t = powertrain.engineData[1].torqueReactionNodes
      if #t > 0 and v.data.nodes[t[1]] ~= nil then
        M.engineNode = t[1]
      end
    end
  end

  local loadedFolder = v.vehicleDirectory..'sounds/'

  --load sbeam files
  local sounds = loadSoundFiles(loadedFolder)

  --no sbeam files on current vehicle, load defaults
  if not sounds then
    loadedFolder = 'vehicles/common/sounds/'
    sounds = loadSoundFiles(loadedFolder)
  end

  if sounds then
    --store in module
    soundBank = sounds

    --check and postprocess them
    for skey, s in pairs(soundBank.sounds) do
      -- set default values
      if s.volumeModifiers == nil then s.volumeModifiers = {} end
      if s.pitchModifiers == nil then s.pitchModifiers= {} end
      if s.profile == nil then s.profile = 'AudioDefaultLoop3D' end

      -- create the sfxprofiles dynamically when filename and profile are specified
      if not s.sfxProfile and s.filename and s.profile then
        -- figure out if the filename was specified relative to the current folder
        s.filename = checkLocalFile(loadedFolder,s.filename)

        -- create the SFXProfile on the T3D - at least the supposed SFXprofilename
        s.sfxProfile = getNextProfile()
        s.waitforloading = 1 -- wait one frame before trying to load the sfxprofile
        s.autocreatedSFXProfile = true
      end

      --try to find our node, default to camera
      s.node = getNodeByName(s.nodeName)
      if s.nodeName == "CAMERA" then s.node = cameraNode end
      if s.nodeName == "ENGINE" then s.node = M.engineNode end
      s.node = s.node or cameraNode -- fall back to camera node

      -- load the sound synchronously with the game engine
      s.clip = createSoundObj(s.filename, s.profile, s.sfxProfile, s.node)
      --log('D', 'sounds.update', 'createSFXSource('..s.sfxProfile..','..s.node..') = '..tostring(s.clip))
      if not s.clip then
        log('W', 'sounds.update', 'unable to create sound, removing it: '..s.sfxProfile)
        soundBank.sounds[skey] = nil
      end
    end

    for k1,snd in pairs(soundBank.sounds) do
      for k2,v2 in pairs(snd) do
        if v2 == "MAXRPM" then snd[k2] = maxrpm end
      end
    end

    --initialize groups
    local soundGroup = v.data.engine and v.data.engine.soundGroup
    for _,vl in pairs(soundBank.sounds) do
      vl.active = (vl.group == "default" or vl.group == soundGroup)
    end

    --initialize per beam sounds
    if v.data.beams then
      for k,bm in pairs(v.data.beams) do
        if bm.soundFile ~= nil then
          local soundTable = {}

          --loop
          local soundProfileType = "AudioDefaultLoop3D"

          --setup our table
          local soundFile = checkLocalFile(v.vehicleDirectory,bm.soundFile)
          soundTable.soundType = bm.soundType
          soundTable.sfxProfile = getNextProfile()
          soundTable.clip = createSoundObj(soundFile, soundProfileType, soundTable.sfxProfile, bm.id1)
          if soundTable.clip then
            soundTable.volumeFactor = bm.volumeFactor or 1
            soundTable.pitchFactor = bm.pitchFactor or 0
            soundTable.decayFactor = bm.decayFactor or 1
            soundTable.decayMode = bm.decayMode or 0   --linear or exponential sound decay?
            soundTable.minStress = bm.minStress or 25000
            soundTable.maxStress = bm.maxStress or 35000

            soundTable.beam = bm.cid
            soundTable.lastStress = 0

            soundTable.smoothing = newTemporalSmoothingNonLinear(10)

            soundTable.resonance = 0
            soundTable.resetTimer = 0

            soundTable.clip:setVolumePitch(0, 0)

            --finally, insert it
            table.insert(beamSounds, soundTable)
          else
            --log('E', 'sounds.init', 'unable to load sound: ' .. tostring(soundFile))
          end
        end
      end
    else
      log('E', 'sounds.init', 'unable to load any sound bank (*.sbeam), that is quite bad :/')
    end
  end

  -- TODO: Find a better place to emit wind sounds. Maybe at the windows?
  if windSound == nil then
    windSound = createSoundObj('art/sound/wind.ogg', 'AudioDefaultLoop3D', "WindTestSound", M.engineNode)
  end

  if wheelsSounds == nil then
    local wheelSoundsNum = 0
    local soundName
    for wi,wd in pairs(wheels.wheels) do
      addWheelSound(wi, wd, "art/sound/rolling_asphalt.ogg", "AudioDefaultLoop3D", "RollingTestSound")
      addWheelSound(wi, wd, "art/sound/skid_asphalt.ogg", "AudioDefaultLoop3D", "SkidTestSound")
    end
  end
end

local function destroy()
  obj:deleteSFXSources()

  soundBank = {}
  beamSounds = {}
  wheelsSounds = nil
  windSound = nil
end


--Siren & horn Mod
local soundCreate = 0
local soundCreateDefault = 0
local isOn = 0
local rumbler = 1
local test
local ambu = 0

--//////////////////////////////////////////

local function loadSounds()
  if soundCreate == 0 then
    metal1 = createSoundObj('art/sound/metal1.ogg', 'AudioDefault3D', "Metal1Sound", M.engineNode) metal1:setVolumePitch(0, 1)
	metal2 = createSoundObj('art/sound/metal2.ogg', 'AudioDefault3D', "Metal2Sound", M.engineNode) metal2:setVolumePitch(0, 1)
	beep = createSoundObj('art/sound/beep.ogg', 'AudioDefaultLoop3D', "BeepSound", M.engineNode) beep:setVolumePitch(0, 1)
    police = createSoundObj('art/sound/policesound.ogg', 'AudioDefaultLoop3D', "PoliceSound", M.engineNode) police:setVolumePitch(0, 1)
	smoke = createSoundObj('art/sound/smoke.ogg', 'AudioDefaultLoop3D', "SmokeSound", M.engineNode) smoke:setVolumePitch(0, 1)
	blinker = createSoundObj('art/sound/blinker.ogg', 'AudioDefaultLoop3D', "BlinkerSound", M.engineNode) blinker:setVolumePitch(0, 1)
	ambulanceSound = createSoundObj('art/sound/ambulancesound.ogg', 'AudioDefaultLoop3D', "AmbulanceSound", M.engineNode) ambulanceSound:setVolumePitch(0, 1)
    policeRumbler1 = createSoundObj('art/sound/rumbler1.ogg', 'AudioDefaultLoop3D', "PoliceSoundRumbler1", M.engineNode) policeRumbler1:setVolumePitch(0, 1)
    policeRumbler2 = createSoundObj('art/sound/rumbler2.ogg', 'AudioDefaultLoop3D', "PoliceSoundRumbler2", M.engineNode) policeRumbler2:setVolumePitch(0, 1)
	ambulanceRumbler = createSoundObj('art/sound/rumbler3.ogg', 'AudioDefaultLoop3D', "ambulanceSoundRumbler", M.engineNode) ambulanceSound:setVolumePitch(0, 1)
   	HornPigeon = createSoundObj('art/sound/horncar.ogg', 'AudioDefaultLoop3D', "HornPigeonSound", M.engineNode) HornPigeon:setVolumePitch(0, 1.5)
	HornSpecial = createSoundObj('art/sound/hornoldcar.ogg', 'AudioDefaultLoop3D', "HornSpecialSound", M.engineNode) HornSpecial:setVolumePitch(0, 1)
	HornTseries = createSoundObj('art/sound/horntruck.ogg', 'AudioDefaultLoop3D', "HornTseriesSound", M.engineNode) HornTseries:setVolumePitch(0, 1)
	HornHseries = createSoundObj('art/sound/horncar.ogg', 'AudioDefaultLoop3D', "HornHseriesSound", M.engineNode) HornHseries:setVolumePitch(0, 0.7)
	HornDefault = createSoundObj('art/sound/horncar.ogg', 'AudioDefaultLoop3D', "HornDefaultSound", M.engineNode) HornDefault:setVolumePitch(0, 1)
	rollgravel = createSoundObj('art/sound/roll_gravel1.ogg', 'AudioDefaultLoop3D', "rollgravel", M.engineNode) rollgravel:setVolumePitch(0, 1)
	HandbrakeLock = createSoundObj('art/sound/handbrakelock.ogg', 'AudioDefault3D', "handbrakeLock", M.engineNode)
	HandbrakeUnlock = createSoundObj('art/sound/handbrakeunlock.ogg', 'AudioDefault3D', "handbrakeUnlock", M.engineNode)
	lightssound = createSoundObj('art/sound/lights.ogg', 'AudioDefault3D', "lights", M.engineNode)
	Shifter1 = createSoundObj('art/sound/shift1.ogg', 'AudioDefault3D', "shift1", M.engineNode)
	Shifter2 = createSoundObj('art/sound/shift2.ogg', 'AudioDefault3D', "shift2", M.engineNode)
	Shifter3 = createSoundObj('art/sound/shift3.ogg', 'AudioDefault3D', "shift3", M.engineNode)
	soundCreate = 1
  end
end

--//////////////////////////////////////////

local function ambulance()
ambu = 1
end

local function pasambulance()
ambu = 0
end

--//////////////////////////////////////////

local function policeOn()
  loadSounds()
  if ambu == 0 then
	if isOn == 0 then
		police:setVolumePitch(1, 1)
		isOn = 1
		M.isOn = isOn
    else
		police:setVolumePitch(0, 1)
		isOn = 0
		M.isOn = isOn
    end
  
  else
	if isOn == 0 then
		ambulanceSound:setVolumePitch(0.7, 1)
		isOn = 1
		M.isOn = isOn
    else
		ambulanceSound:setVolumePitch(0, 1)
		isOn = 0
		M.isOn = isOn
    end
  
  end
end

--//////////////////////////////////////////

local function hornSoundOn()
	
  	M.vehName = readJsonFile("".. v.vehicles.vehicleDirectory .."info.json" or readJsonFile("lua/vehicle/VehicleSettings/parents/defaultInfo.json")  )  
    carName = M.vehName.Name
    carName = carName:gsub("%/", "")
    carName = carName:gsub("% ", "_")
	loadSounds()
		
	if isOn == 0 then
		if carName == "Pigeon" then
			HornPigeon:setVolumePitch(1, 1.5)	
		
			elseif carName == "Special" then
			HornSpecial = createSoundObj('art/sound/hornoldcar.ogg', 'AudioDefaultLoop3D', "HornSpecialSound", M.engineNode) HornSpecial:setVolumePitch(0, 1)
			HornSpecial:setVolumePitch(1, 1)	
			
			elseif carName == "T-Series" then
			HornTseries:setVolumePitch(1, 1)
				
			elseif carName == "H-Series" then
			HornHseries:setVolumePitch(1, 0.7)
					
			else
			HornDefault:setVolumePitch(1, 1)	
	end
	
	elseif isOn == 1 then
		if ambu == 0 then
		police:setVolumePitch(0, 1)
		
		if rumbler == 1 then
		policeRumbler1:setVolumePitch(1, 1)
		rumbler = 2
		
		elseif rumbler == 2 then
		policeRumbler2:setVolumePitch(1, 1)
		rumbler = 1	
		end
		
		elseif ambu == 1 then
		ambulanceSound:setVolumePitch(0, 1)
		ambulanceRumbler:setVolumePitch(0.7, 1)		
		end	
	end	
end

--//////////////////////////////////////////

local function hornSoundOff()
	loadSounds()
	HornPigeon:setVolumePitch(0, 1) 
	HornSpecial:setVolumePitch(0, 1) 
	HornTseries:setVolumePitch(0, 1) 
	HornHseries:setVolumePitch(0, 1) 
	HornDefault:setVolumePitch(0, 1)
	policeRumbler1:setVolumePitch(0, 1)
	policeRumbler2:setVolumePitch(0, 1)
	ambulanceRumbler:setVolumePitch(0, 1)
	if isOn == 1 and ambu == 0 then police:setVolumePitch(1, 1) end
	if isOn == 1 and ambu == 1 then ambulanceSound:setVolumePitch(1, 1) end
end

--//////////////////////////////////////////

local function blinkerOn()
	loadSounds()			
	blinker:setVolumePitch(1, 1)
end

local function blinkerOff()
	loadSounds()	
	blinker:setVolumePitch(0, 1)
end

--//////////////////////////////////////////

local function smokeOn()
	loadSounds()
	smoke:setVolumePitch(1, 1)
end

local function smokeOff()
	loadSounds()
	smoke:setVolumePitch(0, 1)
	beep:setVolumePitch(0, 1)
end

--//////////////////////////////////////////

local function beep()
	loadSounds()
	beep:setVolumePitch(1, 1)
end

--//////////////////////////////////////////
--Siren & horn Mod

-- this function enables or disables the data reporting to the UI. It is quite performance heavy and should be only used for debugging
local function setUIDebug(enabled, data)
  M.uiDebugging = enabled
  -- todo: save data
end

local function onDeserialized()
end

-- public interface
--Siren & horn Mod
M.blinkerOn = blinkerOn
M.smokeOn = smokeOn
M.beep = beep
M.smokeOff = smokeOff
M.blinkerOff = blinkerOff
M.pasambulance = pasambulance
M.loadSounds = loadSounds
M.ambulance = ambulance
M.policeOff = policeOff
M.policeOn = policeOn
M.hornSoundOff = hornSoundOff
M.hornSoundOn = hornSoundOn
--Siren & horn Mod

M.destroy = destroy
M.init = init
M.setUIDebug = setUIDebug
M.createSFXSource = createSFXSource
M.onDeserialized = onDeserialized --this enables serialization of all M. values for the module so they survive reloads
return M
